"""
The Graph4NLP library uses the class :py:class:`GraphData` as the representation for structured data (graphs).
:py:class:`GraphData` supports basic operations to the graph, including adding nodes and edges. :py:class:`GraphData` also
supports adding features which are in tensor form, and attributes which are of arbitrary form to the corresponding
nodes or edges. Batching operations is also supported by :py:class:`GraphData`.

"""
import os
"""
Log level: 0 for verbose, 1 for warnings only, 2 for muted. Default is 0.
"""
log_level = os.environ.get("G4NLP_LOG_LEVEL")
if log_level is None:
    log_level = 0

import warnings
from collections import namedtuple

import dgl
import scipy.sparse
import torch
from torch.nn.utils.rnn import pad_sequence

from .utils import SizeMismatchException, EdgeNotFoundException
from .utils import check_and_expand, int_to_list, entail_zero_padding, slice_to_list
from .views import NodeView, NodeFeatView, EdgeView, BatchNodeFeatView, BatchEdgeFeatView

EdgeIndex = namedtuple('EdgeIndex', ['src', 'tgt'])

node_feature_factory = dict
node_attribute_factory = list
single_node_attr_factory = dict
res_init_node_attr = {'node_attr': None}
res_init_node_features = {'node_feat': None, 'node_emb': None}

eid_nids_mapping_factory = dict
nids_eid_mapping_factory = dict
edge_feature_factory = dict
edge_attribute_factory = list
single_edge_attr_factory = dict
res_init_edge_features = {'edge_feat': None, 'edge_emb': None, 'edge_weight': None}
res_init_edge_attributes = {'edge_attr': None}

graph_data_factory = dict


class GraphData(object):
    """
    Represent a single graph with additional attributes.
    """

    def __init__(self, src=None, device=None):
        """
        Parameters
        ----------
        src: GraphData, default=None
            The source graph. If not None, then the newly generated graph is a copy of :py:class:`src`.
        device: str, default=None
            The device descriptor for graph. By default it is None.
        """

        # Initialize internal data storages.
        self._node_attributes = node_attribute_factory()
        self._node_features = node_feature_factory(res_init_node_features)
        self._edge_indices = EdgeIndex(src=[], tgt=[])
        self._nids_eid_mapping = nids_eid_mapping_factory()
        self._edge_features = edge_feature_factory(res_init_edge_features)
        self._edge_attributes = edge_attribute_factory()
        self.graph_attributes = graph_data_factory()
        self.device = device

        # Batch information. If this instance is not a batch, then the following attributes are all `None`.
        self._is_batch = False  # Bool flag indicating whether this graph is a batch graph
        self.batch = None  # Batch node indices
        self.batch_size = None  # Batch size
        self._batch_num_nodes = None  # Subgraph node number list with the length of batch size
        self._batch_num_edges = None  # Subgraph edge number list with the length of batch size

        if src is not None:
            if isinstance(src, GraphData):
                self.from_graphdata(src)
            else:
                raise NotImplementedError

    def to(self, device):
        """
        Move the GraphData object to different devices(cpu, gpu, etc.). The usage of this method is similar to that of
        torch.Tensor and dgl.DGLGraph

        Parameters
        ----------
        device: str
            The target device.

        Returns
        -------
        self
        """
        self.device = device
        for k, v in self._node_features.items():
            if isinstance(v, torch.Tensor):
                self._node_features[k] = v.to(device)
        for k, v in self._edge_features.items():
            if isinstance(v, torch.Tensor):
                self._edge_features[k] = v.to(device)
        return self

    # Node operations
    @property
    def nodes(self) -> NodeView:
        """
        Return a node view through which the user can access the features and attributes. A NodeView object provides a
        high-level view of the underlying storage of the features and supports both query and modification to the original
        storage.

        Returns
        -------
        node: NodeView
            The node view
        """
        return NodeView(self)

    def get_node_num(self) -> int:
        """
        Get the number of nodes in the graph.

        Returns
        -------
        num_nodes: int
            The number of nodes in the graph.
        """
        return len(self._node_attributes)

    def add_nodes(self, node_num: int) -> None:
        """
        Add a number of nodes to the graph.

        Parameters
        ------
        node_num: int
            The number of nodes to be added
        """
        assert node_num > 0, "The number of nodes to be added should be greater than 0. (Got {})".format(node_num)

        # Create placeholders in the node attribute dictionary
        self._node_attributes.extend([single_node_attr_factory(**res_init_node_attr) for _ in range(node_num)])

        # Do padding in the node feature dictionary
        for key in self._node_features.keys():
            self._node_features[key] = entail_zero_padding(self._node_features[key], node_num)

    # Node feature operations
    @property
    def node_features(self) -> NodeFeatView:
        """
        Access and modify node feature vectors (tensor).
        This property can be accessed in a dict-of-dict fashion, with the order being [name][index].
        'name' indicates the name of the feature vector. 'index' selects the specific nodes to be accessed.
        When accessed independently, returns the feature dictionary with the format {name: tensor}

        Examples
        --------
        >>> g = GraphData()
        >>> g.add_nodes(10)
        >>> import torch
        >>> g.node_features['x'] = torch.rand((10, 10))
        >>> g.node_features['x'][0]
        torch.Tensor([0.1036, 0.6757, 0.4702, 0.8938, 0.6337, 0.3290, 0.6739, 0.1091, 0.7996, 0.0586])

        Returns
        -------
        NodeFeatView
        """

        return self.nodes[:].features

    def get_node_features(self, nodes: int or slice) -> torch.Tensor:
        """
        Get the node feature dictionary of the `nodes`

        Parameters
        ----------
        nodes: int or slice
            The nodes to be accessed

        Returns
        -------
        node_features: dict
            The reference dict of the actual tensor
        """
        ret = dict()
        for key in self._node_features.keys():
            if self._node_features[key] is None:
                ret[key] = None
            else:
                ret[key] = self._node_features[key][nodes]
        return ret

    def node_feature_names(self):
        """
        Get the names of node features.

        Returns
        -------
        Union[str]:
            The collection of feature names.
        """
        return self._node_features.keys()

    def set_node_features(self, nodes: int or slice, new_data: dict) -> None:
        """
        Set the features of the `nodes` with the given `new_data``.

        Parameters
        ----------
        nodes: int or slice
            The nodes involved
        new_data: dict
            The new data to write. Key indicates feature name and value indicates the actual value

        Raises
        ----------
        SizeMismatchException
            If the size of the new features does not match the node number
        """
        # Consistency check
        for key in new_data.keys():
            if key not in self._node_features or self._node_features[key] is None:  # A new feature is added
                # If the shape of the new feature does not match the number of existing nodes, then error occurs
                if (not isinstance(nodes, slice)) or (
                        len(slice_to_list(nodes, self.get_node_num())) != self.get_node_num()):
                    raise ValueError(
                        'The new feature `{}\' should cover all existing {} nodes!'.format(key, self.get_node_num()))

        # Modification
        for key, value in new_data.items():
            # Node-shape check
            assert value.shape[0] == self.get_node_num(), \
                "The shape feature '{}' does not match the number of nodes in the graph. Got a {} tensor but have {} nodes.".format(
                    key, value.shape, self.get_node_num())

            assert isinstance(value, torch.Tensor), "`{}' is not a tensor. Node features are expected to be tensor."

            value_on_device = value
            if key not in self._node_features or self._node_features[key] is None:
                self._node_features[key] = value_on_device
            else:
                if nodes == slice(None, None, None):
                    self._node_features[key] = value_on_device
                else:
                    self._node_features[key][nodes] = value_on_device

    # Node attribute operations
    @property
    def node_attributes(self) -> list:
        """
        Access node attribute dictionary

        Returns
        -------
        node_attribute_dict: dict
            The dict of node attributes
        """
        return self._node_attributes

    def get_node_attrs(self, nodes: int or slice):
        """
        Get the attributes of the given `nodes`.

        Parameters
        ----------
        nodes: int
         or slice
            The given node index

        Returns
        -------
        list
            The node attribute dictionary.
        """
        return self._node_attributes[nodes]

    # Edge views and operations
    @property
    def edges(self):
        """
        Return an edge view of the edges and the corresponding data

        Returns
        -------
        edges: EdgeView
        """
        return EdgeView(self)

    def get_edge_num(self) -> int:
        """
        Get the number of edges in the graph

        Returns
        -------
        num_edges: int
            The number of edges
        """
        return len(self._edge_indices.src)

    def add_edge(self, src: int, tgt: int):
        """
        Add one edge to the graph.

        Parameters
        ----------
        src: int
            Source node index
        tgt: int
            Target node index

        Raises
        ------
        ValueError
            If one of the endpoints of the edge doesn't exist in the graph.
        """
        # Consistency check
        if (src < 0 or src >= self.get_node_num()) and (tgt < 0 and tgt >= self.get_node_num()):
            raise ValueError('Endpoint not in the graph.')

        # Duplicate edge check. If the edge to be added already exists in the graph, then skip it.
        endpoint_tuple = (src, tgt)
        if endpoint_tuple in self._nids_eid_mapping.keys():
            if log_level < 2:
                warnings.warn('Edge {} is already in the graph. Skipping this edge.'.format(endpoint_tuple), Warning)
            return

        # Append to the mapping list
        eid = self.get_edge_num()
        self._nids_eid_mapping[endpoint_tuple] = eid

        # Add edge
        self._edge_indices.src.append(src)
        self._edge_indices.tgt.append(tgt)

        # Initialize edge feature and attribute
        # 1. create placeholder in edge attribute dictionary
        self._edge_attributes.append(single_edge_attr_factory(**res_init_edge_attributes))
        # 2. perform zero padding
        for key in self._edge_features.keys():
            self._edge_features[key] = entail_zero_padding(self._edge_features[key], 1)

    def add_edges(self, src: int or list, tgt: int or list):
        """
        Add a bunch of edges to the graph.

        Parameters
        ----------
        src: int or list
            Source node indices
        tgt: int or list
            Target node indices

        Raises
        ------
        ValueError
            If the lengths of `src` and `tgt` don't match or one of the list contains no element.
        """
        src, tgt = check_and_expand(int_to_list(src), int_to_list(tgt))
        assert len(src) == len(tgt), "Length of the source and target indices is not the same. " \
                                     "Got {} source nodes and {} target nodes".format(len(src), len(tgt))
        for src_idx, tgt_idx in zip(src, tgt):
            # Consistency check
            if (src_idx < 0 or src_idx >= self.get_node_num()) and (tgt_idx < 0 and tgt_idx >= self.get_node_num()):
                raise ValueError('Endpoint not in the graph.')

        current_num_edges = len(self._edge_attributes)
        duplicate_edge_indices = list()
        for i in range(len(src)):
            # Duplicate edge check. If the edge to be added already exists in the graph, then skip it.
            endpoint_tuple = (src[i], tgt[i])
            if endpoint_tuple in self._nids_eid_mapping.keys():
                if log_level < 2:
                    warnings.warn('Edge {} is already in the graph. Skipping this edge.'.format(endpoint_tuple), Warning)
                duplicate_edge_indices.append(i)
                continue
            self._nids_eid_mapping[endpoint_tuple] = current_num_edges + i

        # Remove duplicate edges
        duplicate_edge_indices.reverse()  # Needs to be reversed first to avoid index overflow after popping.
        for edge_index in duplicate_edge_indices:
            src.pop(edge_index)
            tgt.pop(edge_index)

        num_edges = len(src)

        # Add edge indices
        self._edge_indices.src.extend(src)
        self._edge_indices.tgt.extend(tgt)

        # Initialize edge attributes and features
        self._edge_attributes.extend([single_edge_attr_factory(**res_init_edge_attributes) for _ in range(num_edges)])
        for key in self._edge_features.keys():
            self._edge_features[key] = entail_zero_padding(self._edge_features[key], num_edges)

    def edge_ids(self, src: int or list, tgt: int or list) -> list:
        """
        Convert the given endpoints to edge indices.

        Parameters
        ----------
        src: int or list
            The index of source node(s).
        tgt: int or list
            The index of target node(s).

        Returns
        -------
        list
            The index of corresponding edges.

        Raises
        ------
        EdgeNotFoundException
            If the edge is not in the graph.
        """
        assert isinstance(src, int) or isinstance(src, list), "`src` should be either int or list."
        assert isinstance(tgt, int) or isinstance(tgt, list), "`tgt` should be either int or list."
        src, tgt = check_and_expand(int_to_list(src), int_to_list(tgt))
        eid_list = []
        try:
            for src_idx, tgt_idx in zip(src, tgt):
                eid_list.append(self._nids_eid_mapping[(src_idx, tgt_idx)])
        except KeyError:
            raise EdgeNotFoundException('Edge {} does not exist!'.format((src, tgt)))
        return eid_list

    def get_all_edges(self) -> list:
        """
        Get all the edges in the graph

        Returns
        -------
        edges: list
            List of edges. Each edge is in the shape of the endpoint tuple (src, dst).
        """
        edges = []
        for i in range(self.get_edge_num()):
            edges.append((self._edge_indices.src[i], self._edge_indices.tgt[i]))
        return edges

    # Edge feature operations
    @property
    def edge_features(self):
        """
        Get all the edge features in a dictionary.
        Returns
        -------
        dict:
            Edge features with the keys being the feature names and values be the corresponding tensors.
        """
        return self.edges[:].features

    def remove_all_edges(self):
        """
        Remove all the edges and the corresponding features and attributes in GraphData.

        Examples
        --------
        >>> g = GraphData()
        >>> g.add_nodes(10)
        >>> g.add_edges(list(range(0, 9, 1)), list(range(1, 10, 1)))
        >>> g.edge_features['random'] = torch.rand((9, 1024, 1024))     # Added some feature tensors to the edges
        >>> g.remove_all_edges()    # Remove all edges and the corresponding data. The tensor memory is freed now.

        Returns
        -------
        None
        """

        self._edge_indices = EdgeIndex(src=[], tgt=[])
        self._nids_eid_mapping = nids_eid_mapping_factory()
        self._edge_features = edge_feature_factory(res_init_edge_features)
        self._edge_attributes = edge_attribute_factory()

    def get_edge_feature(self, edges: list):
        """
        Get the feature of the given edges.

        Parameters
        ----------
        edges: list
            Edge indices

        Returns
        -------
        dict
            The dictionary containing all relevant features.
        """
        ret = {}
        for key in self._edge_features.keys():
            if self._edge_features[key] is None:
                ret[key] = None
            else:
                ret[key] = self._edge_features[key][edges]
        return ret

    def get_edge_feature_names(self):
        """Get all the names of edge features"""
        return self._edge_features.keys()

    def set_edge_feature(self, edges: int or slice or list, new_data: dict):
        """
        Set edge feature

        Parameters
        ----------
        edges: int or list or slice
            Edge indices
        new_data: dict
            New data

        Raises
        ----------
        SizeMismatchException
            If the size of the new features does not match the node number
        """
        # Consistency check
        for key in new_data.keys():
            if key not in self._edge_features or self._edge_features[key] is None:  # A new feature is added
                # If the shape of the new feature does not match the number of existing nodes, then error occurs
                if (not isinstance(edges, slice)) or (
                        len(slice_to_list(edges, self.get_edge_num())) != self.get_edge_num()):
                    raise SizeMismatchException(
                        'The new feature `{}\' should cover all existing {} edges!'.format(key, self.get_edge_num()))

        # Modification
        for key, value in new_data.items():
            assert isinstance(value, torch.Tensor), "`{}' is not a tensor. Node features are expected to be tensor."
            assert value.shape[0] == self.get_edge_num(), "Length of the feature vector does not match the edge number." \
                                                          "Got tensor '{}' of shape {} but the graph has only {} edges.".format(
                key, value.shape, self.get_edge_num())
            # Move the new value to the device consistent with current graph
            value_on_device = value.to(self.device)

            if key not in self._edge_features or self._edge_features[key] is None:
                self._edge_features[key] = value_on_device
            elif edges == slice(None, None, None):
                # Same as node features, if the edges to be modified is all the edges in the graph.
                self._edge_features[key] = value_on_device
            else:
                self._edge_features[key][edges] = value_on_device

    # Edge attribute operations
    @property
    def edge_attributes(self):
        """
        Get the edge attributes in a list.
        Returns
        -------
        list:
            A list of dictionaries. Each dictionary represents all the attributes on the corresponding edge.
        """
        return self._edge_attributes

    # Conversion utility functions
    def to_dgl(self) -> dgl.DGLGraph:
        """
        Convert to dgl.DGLGraph
        Note that there will be some information loss when calling this funciton, e.g. the batch-related information
        will not be copied to DGLGraph since it is only intended for computation.

        Returns
        -------
        g: dgl.DGLGraph
            The converted dgl.DGLGraph
        """
        dgl_g = dgl.DGLGraph().to(self.device)
        # Add nodes and their features
        dgl_g.add_nodes(num=self.get_node_num())
        for key, value in self._node_features.items():
            if value is not None:
                dgl_g.ndata[key] = value
        # Add edges and their features
        dgl_g.add_edges(u=self._edge_indices.src, v=self._edge_indices.tgt)
        for key, value in self._edge_features.items():
            if value is not None:
                dgl_g.edata[key] = value
        return dgl_g

    def from_dgl(self, dgl_g: dgl.DGLGraph):
        """
        Build the graph from dgl.DGLGraph

        Parameters
        ----------
        dgl_g: dgl.DGLGraph
            The source graph
        """
        assert self.get_edge_num() == 0 and self.get_node_num() == 0, \
            'This graph isn\'t an empty graph. Please use an empty graph for conversion.'

        # Add nodes
        self.add_nodes(dgl_g.number_of_nodes())
        for k, v in dgl_g.ndata.items():
            self.node_features[k] = v

        # Add edges
        src_tensor, tgt_tensor = dgl_g.edges()
        src_list = list(src_tensor.detach().cpu().numpy())
        tgt_list = list(tgt_tensor.detach().cpu().numpy())
        self.add_edges(src_list, tgt_list)
        for k, v in dgl_g.edata.items():
            self.edge_features[k] = v
        return self

    def from_dense_adj(self, adj: torch.Tensor):
        """
        Construct a graph from a dense (2-D NxN) adjacency matrix with the edge weights represented by the value of
        the matrix entries
        Parameters
        ----------
        adj: torch.Tensor
            The tensor representing the adjacency matrix.

        Returns
        -------
        self
        """
        assert adj.dim() == 2, 'Adjancency matrix is not 2-dimensional.'
        assert adj.shape[0] == adj.shape[1], 'Adjancecy is not a square.'

        node_num = adj.shape[0]
        self.add_nodes(node_num)
        edge_weight = []
        for i in range(adj.shape[0]):
            for j in range(adj.shape[1]):
                if adj[i][j] != 0:
                    self.add_edge(i, j)
                    edge_weight.append(adj[i][j])
        edge_weight = torch.stack(edge_weight, dim=0)
        self.edge_features['edge_weight'] = edge_weight
        return self

    def from_scipy_sparse_matrix(self, adj: scipy.sparse.coo_matrix):
        """
        Construct a graph from a sparse adjacency matrix with the edge weights represented by the value of
        the matrix entries
        Parameters
        ----------
        adj: scipy.sparse.coo_matrix
            The object representing the sparse adjacency matrix.

        Returns
        -------
        self
        """
        assert adj.shape[0] == adj.shape[1], 'Got an adjancecy matrix which is not a square.'

        num_nodes = adj.shape[0]
        self.add_nodes(num_nodes)

        for i in range(adj.row.shape[0]):
            self.add_edge(adj.row[i], adj.col[i])
        self.edge_features['edge_weight'] = torch.tensor(adj.data)
        return self

    def adj_matrix(self, batch_view=False, post_processing_fn=None):
        """
        Returns the adjacency matrix of the graph. Returns a 2D tensor if it is a single graph and a 3D tensor if it
        is a batched graph, with the matrices padded with 0 (B x N x N)

        Parameters
        ----------
        batch_view: bool
            Whether to return a batched view of the adjacency matrix(3D, True) or not (2D).
        post_processing_fn: function
            A callback function which takes a binary adjacency matrix (2D) and do some post-processing on it.
            The return of this function should also be N x N

        Returns
        -------
        torch.Tensor:
            The adjacency matrix (N x N if batch_view=False and B x N x N if batch_view=True).
        """
        if batch_view is False:
            ret = torch.zeros((self.get_node_num(), self.get_node_num())).to(self.device)
            all_edges = self.edges()
            for i in range(len(all_edges)):
                u, v = all_edges[i]
                ret[u][v] = 1
            if post_processing_fn is not None:
                ret = post_processing_fn(ret)
            return ret
        else:
            assert self._is_batch, "Cannot enable batch view on a non-batch graph!"
            max_num_nodes = max(self._batch_num_nodes)
            ret = torch.zeros((self.batch_size, max_num_nodes, max_num_nodes)).to(self.device)
            cum_num_nodes = 0
            cum_num_edges = 0
            all_edges = self.get_all_edges()
            for i in range(self.batch_size):
                edges = all_edges[cum_num_edges:cum_num_edges+self._batch_num_edges[i]]
                for edge in edges:
                    ret[i][edge[0] - cum_num_nodes, edge[1] - cum_num_nodes] = 1
                if post_processing_fn is not None:
                    ret[i] = post_processing_fn(ret[i])
                cum_num_nodes += self._batch_num_nodes[i]
                cum_num_edges += self._batch_num_edges[i]
            return ret

    def sparse_adj(self, batch_view=False):
        """
        Return the scipy.sparse.coo_matrix form of the adjacency matrix
        Parameters
        ----------
        batch_view: bool
            Whether to return the split view of the adjacency matrix. Return a list of COO matrix if True

        Returns
        -------
        torch.sparse_coo_tensor or list of torch.sparse_coo_tensor
        """
        row = torch.tensor(self._edge_indices[0]).to(self.device)
        col = torch.tensor(self._edge_indices[1]).to(self.device)
        data = torch.ones(self.get_edge_num()).to(self.device)
        if not batch_view:
            indices = torch.stack([row, col])
            matrix = torch.sparse_coo_tensor(indices=indices, values=data,
                                             size=(self.get_node_num(), self.get_node_num()))
            return matrix
        else:
            if self._is_batch is not True:
                raise Exception("Cannot enable batch view of COO adjacency matrix on a non-batch graph.")
            matrices = []
            cum_num_edges = 0
            cum_num_nodes = 0
            for i in range(self.batch_size):
                num_edges = self._batch_num_edges[i]
                num_nodes = self._batch_num_nodes[i]
                # Slicing the matrix one by one
                cur_row = row[cum_num_edges:cum_num_edges + num_edges]
                cur_col = col[cum_num_edges:cum_num_edges + num_edges]
                cur_data = data[cum_num_edges:cum_num_edges + num_edges]
                cur_row -= cum_num_nodes
                cur_col -= cum_num_nodes
                indices = torch.stack([cur_row, cur_col])
                cur_matrix = torch.sparse_coo_tensor(indices=indices, values=cur_data, size=(num_nodes, num_nodes))
                matrices.append(cur_matrix)
                cum_num_edges += num_edges
                cum_num_nodes += num_nodes
            return matrices

    def from_graphdata(self, src):
        """Build a clone from a source GraphData"""

        # Add nodes and edges
        self.add_nodes(src.get_node_num())
        self.add_edges(src._edge_indices.src, src._edge_indices.tgt)

        # Deepcopy of feature tensors
        for k, v in src._node_features.items():
            self._node_features[k] = v
        for k, v in src._edge_features.items():
            self._edge_features[k] = v

        # Copy attributes
        import copy
        self._node_attributes = copy.deepcopy(src.node_attributes)
        self._edge_attributes = copy.deepcopy(src.edge_attributes)
        self.graph_attributes = copy.deepcopy(src.graph_attributes)

        # Copy batch information if necessary
        if src._is_batch:
            self.copy_batch_info(src)

        # Move data to the device of the source graph
        self.to(src.device)

    def copy_batch_info(self, batch):
        """
        Copy all the information related to the batching.
        Parameters
        ----------
        batch:
            The source batch from which the information comes.

        Returns
        -------
        None
        """
        self._is_batch = True
        self.batch = batch.batch
        self.device = batch.device
        self.batch_size = batch.batch_size
        self._batch_num_edges = batch._batch_num_edges
        self._batch_num_nodes = batch._batch_num_nodes

    @property
    def batch_node_features(self):
        """
        Get a view of the batched(padded) version of the node features. Shape: (B, N, D)

        Returns
        -------
        BatchNodeFeatView
        """
        return BatchNodeFeatView(self)

    def _get_batch_node_features(self, item=None):
        """
        Get the batched view of node feature tensors, i.e., tensors in (B, N, D) view

        Parameters
        -------
        item: str
            The name of the features. If None then return a dictionary of all the features.

        Returns
        -------
        dict or tensor:
            A dictionary containing the node feature names and the corresponding batch-view tensors, or just the
            specified tensor.
        """
        if not self._is_batch:
            raise Exception("Calling batch_node_features() method on a non-batch graph.")
        if item is None:
            batch_node_features = dict()
            separate_features = self.split_node_features
            for k, v in separate_features.items():
                batch_node_features[k] = pad_sequence(list(v), batch_first=True)
            return batch_node_features
        else:
            if (item not in self.node_features.keys()) or (self.node_features[item] is None):
                raise Exception("Node feature {} doesn't exist!".format(item))
            return pad_sequence(self.split_node_features[item], batch_first=True)

    def _set_batch_node_features(self, key, value):
        """
        Set node features in batch view.

        Parameters
        ----------
        key: str
            The name of the feature.
        value: Tensor
            The values to be written, in the shape of (B, N, D)
        """
        individual_features = [value[i, :self._batch_num_nodes[i]] for i in range(len(self._batch_num_nodes))]
        self.set_node_features(slice(None, None, None), {key: torch.cat(individual_features)})

    @property
    def batch_edge_features(self):
        """
        Edge version of self.batch_node_features

        Returns
        -------
        BatchEdgeFeatView
        """
        return BatchEdgeFeatView(self)

    def _get_batch_edge_features(self, item=None):
        """
        An edge version of :py:method `batch_node_features`.

        Returns
        -------
        dict or tensor:
            A dictionary containing the edge feature names and the corresponding batch-view tensors.
        """
        if not self._is_batch:
            raise Exception("Calling batch_edge_features() method on a non-batch graph.")
        if item is None:
            batch_edge_features = dict()
            for k, v in self.split_edge_features.items():
                batch_edge_features[k] = pad_sequence(list(v), batch_first=True)
            return batch_edge_features
        else:
            if (item not in self.edge_features.keys()) or (self.edge_features[item] is None):
                raise Exception("Edge feature {} doesn't exist!".format(item))
            return pad_sequence(self.split_edge_features[item], batch_first=True)

    def _set_batch_edge_features(self, key, value):
        individual_features = [value[i, :self._batch_num_edges[i]] for i in range(len(self._batch_num_edges))]
        self.set_edge_feature(slice(None, None, None), {key: torch.cat(individual_features)})

    @property
    def split_node_features(self):
        if not self._is_batch:
            raise Exception("Calling split_node_features() method on a non-batch graph.")
        node_features = dict()
        for feature in self.node_features.keys():
            if self.node_features[feature] is None:
                continue
            node_features[feature] = torch.split(self.node_features[feature],
                                                 split_size_or_sections=self._batch_num_nodes)
        return node_features

    @property
    def split_edge_features(self):
        if not self._is_batch:
            raise Exception("Calling split_edge_features() method on a non-batch graph.")
        edge_features = dict()
        for feature in self.edge_features.keys():
            if self.edge_features[feature] is None:
                continue
            edge_features[feature] = torch.split(self.edge_features[feature],
                                                 split_size_or_sections=self._batch_num_edges)
        return edge_features

    def split_features(self, input_tensor: torch.Tensor, type='node') -> torch.Tensor:
        """
        Convert a tensor from [N, *] to [B, N_max, *] with zero padding according to the batch information stored in
        the graph.
        Parameters
        ----------
        input_tensor: torch.Tensor
        The original tensor to be split.
        type: str
        'node' or 'edge'. Indicates the source of batch information.
        Returns
        -------
        torch.Tensor
        The split tensor.
        """
        input_tensor = input_tensor.to(self.device)
        assert self._is_batch, "Cannot invoke `batch_split` method on a non-batch graph."
        if type == 'node':
            info_src = self._batch_num_nodes
        elif type == 'edge':
            info_src = self._batch_num_edges
        else:
            raise NotImplementedError("Currently only 'node' and 'edge' is accepted in GraphData.split_features().")
        num_instance = 0
        for number in info_src:
            num_instance += number
        assert num_instance == input_tensor.shape[0], "Number of instances " \
                                                      "doesn't match: The graph " \
                                                      "has {} instances while the input " \
                                                      "contains {}".format(num_instance, input_tensor.shape[0])
        n_max = max(info_src)
        output = torch.zeros(size=(self.batch_size, n_max, *input_tensor.shape[1:])).to(self.device)
        split_input = torch.split(tensor=input_tensor, split_size_or_sections=info_src)
        for i in range(self.batch_size):
            output[i, :info_src[i]] = split_input[i]
        return output


def from_dgl(g: dgl.DGLGraph) -> GraphData:
    """
    Convert a dgl.DGLGraph to a GraphData object.

    Parameters
    ----------
    g: dgl.DGLGraph
        The source graph in DGLGraph format.

    Returns
    -------
    GraphData
        The converted graph in GraphData format.
    """
    graph = GraphData()
    graph.from_dgl(g)
    return graph


def to_batch(graphs: list = None) -> GraphData:
    """
    Convert a list of GraphData to a large graph (a batch).

    Parameters
    ----------
    graphs: list of GraphData
        The list of GraphData to be batched

    Returns
    -------
    GraphData
        The large graph containing all the graphs in the batch.
    """

    # Check
    assert isinstance(graphs, list), "to_batch() only accepts list of GraphData!"
    assert len(graphs) > 0, "Cannot convert an empty list of graphs into a big batched graph!"

    # Optimized version
    big_graph = GraphData()
    big_graph._is_batch = True
    big_graph.device = graphs[0].device

    total_num_nodes = 0
    for g in graphs:
        total_num_nodes += g.get_node_num()

    # Step 1: Add nodes
    big_graph.add_nodes(total_num_nodes)

    # Step 2: Set node features
    node_features = dict()
    for g in graphs:
        for feature_name in g.node_features.keys():
            if feature_name in node_features:
                node_features[feature_name].append(g.node_features[feature_name])
            else:
                node_features[feature_name] = [g.node_features[feature_name]]
    for k, v in node_features.items():
        if None in v:
            continue
        else:
            feature_tensor = torch.cat(v, dim=0)
            big_graph.node_features[k] = feature_tensor

    # Step 3: Set node attributes
    total_node_count = 0
    for g in graphs:
        for i in range(g.get_node_num()):
            big_graph.node_attributes[total_node_count] = g.node_attributes[i]
            total_node_count += 1

    # Step 4: Add edges
    def stack_edge_indices(gs):
        all_edge_indices = EdgeIndex(src=[], tgt=[])
        cumulative_node_num = 0
        for g in gs:
            for edge_index_tuple in g.get_all_edges():
                src, tgt = edge_index_tuple
                src += cumulative_node_num
                tgt += cumulative_node_num
                all_edge_indices.src.append(src)
                all_edge_indices.tgt.append(tgt)
            cumulative_node_num += g.get_node_num()
        return all_edge_indices

    all_edge_indices = stack_edge_indices(graphs)
    big_graph.add_edges(all_edge_indices.src, all_edge_indices.tgt)

    # Step 5: Add edge features
    edge_features = dict()
    for g in graphs:
        for feature_name in g.edge_features.keys():
            if feature_name in edge_features:
                edge_features[feature_name].append(g.edge_features[feature_name])
            else:
                edge_features[feature_name] = [g.edge_features[feature_name]]
    for k, v in edge_features.items():
        if None in v:
            continue
        else:
            feature_tensor = torch.cat(v, dim=0)
            big_graph.edge_features[k] = feature_tensor

    # Step 6: Add edge attributes
    total_edge_count = 0
    for g in graphs:
        for i in range(g.get_edge_num()):
            big_graph.edge_attributes[total_edge_count] = g.edge_attributes[i]
            total_edge_count += 1

    # Step 7: Batch information preparation
    big_graph.batch_size = len(graphs)
    batch_numbers = []
    for i in range(len(graphs)):
        batch_numbers.extend([i] * graphs[i].get_node_num())
    big_graph.batch = batch_numbers
    big_graph._batch_num_nodes = [g.get_node_num() for g in graphs]
    big_graph._batch_num_edges = [g.get_edge_num() for g in graphs]

    return big_graph


def from_batch(batch: GraphData) -> list:
    """
    Convert a batch consisting of several GraphData instances to a list of GraphData instances.

    Parameters
    ----------
    batch: GraphData
        The source batch to be split.

    Returns
    -------
    list
        A list containing all the GraphData instances contained in the source batch.
    """

    num_nodes = batch._batch_num_nodes
    num_edges = batch._batch_num_edges
    all_edges = batch.get_all_edges()
    batch_size = batch.batch_size
    ret = []
    cum_n_nodes = 0
    cum_n_edges = 0

    # Construct graph respectively
    for i in range(batch_size):
        g = GraphData(device=batch.device)
        g.add_nodes(num_nodes[i])
        edges = all_edges[cum_n_edges:cum_n_edges + num_edges[i]]
        src, tgt = [e[0] - cum_n_nodes for e in edges], [e[1] - cum_n_nodes for e in edges]
        g.add_edges(src, tgt)
        cum_n_edges += num_edges[i]
        cum_n_nodes += num_nodes[i]
        ret.append(g)

    # Add node and edge features

    for k, v in batch._node_features.items():
        if v is not None:
            cum_n_nodes = 0  # Cumulative node numbers
            for i in range(batch_size):
                ret[i].node_features[k] = v[cum_n_nodes:cum_n_nodes + num_nodes[i]]
                cum_n_nodes += num_nodes[i]

    for k, v in batch._edge_features.items():
        if v is not None:
            cum_n_edges = 0  # Cumulative edge numbers
            for i in range(batch_size):
                ret[i].edge_features[k] = v[cum_n_edges:cum_n_edges + num_edges[i]]
                cum_n_edges += num_edges[i]

    cum_n_nodes = 0
    cum_n_edges = 0

    # Add node and edge attributes
    for graph_cnt in range(batch_size):
        for num_graph_nodes in range(num_nodes[graph_cnt]):
            ret[graph_cnt].node_attributes[num_graph_nodes] = batch.node_attributes[cum_n_nodes + num_graph_nodes]
        for num_graph_edges in range(num_edges[graph_cnt]):
            ret[graph_cnt].edge_attributes[num_graph_edges] = batch.edge_attributes[cum_n_edges + num_graph_edges]
        cum_n_edges += num_edges[graph_cnt]
        cum_n_nodes += num_nodes[graph_cnt]

    return ret


# Testing code
if __name__ == '__main__':
    a1 = GraphData()
    a1.add_nodes(10)
    a2 = GraphData()
    a2.add_nodes(15)
    a3 = to_batch([a1, a2])
    print(len(a1.node_attributes), len(a2.node_attributes), len(a3.node_attributes))



import copy
import enum
import json

import torch
from stanfordcorenlp import StanfordCoreNLP

from .base import StaticGraphConstructionBase
from ...data.data import GraphData, to_batch


class DependencyBasedGraphConstruction(StaticGraphConstructionBase):
    """
        Dependency-parsing-tree based graph construction class

    Parameters
    ----------
    embedding_style: dict
        Specify embedding styles including ``single_token_item``, ``emb_strategy``, ``num_rnn_layers``, ``bert_model_name`` and ``bert_lower_case``.
    vocab: VocabModel
        Vocabulary including all words appeared in graphs.
    """

    def __init__(self, embedding_style, vocab, hidden_size=300, fix_word_emb=True, fix_bert_emb=True, word_dropout=None,
                 rnn_dropout=None):
        super(DependencyBasedGraphConstruction, self).__init__(word_vocab=vocab,
                                                               embedding_styles=embedding_style,
                                                               hidden_size=hidden_size,
                                                               fix_word_emb=fix_word_emb,
                                                               fix_bert_emb=fix_bert_emb,
                                                               word_dropout=word_dropout,
                                                               rnn_dropout=rnn_dropout)
        self.vocab = vocab
        self.verbase = 1

    def add_vocab(self, g):
        """
            Add node tokens appeared in graph g to vocabulary.

        Parameters
        ----------
        g: GraphData
            Graph data-structure.

        """
        for i in range(g.get_node_num()):
            attr = g.get_node_attrs(i)[i]
            self.vocab.word_vocab._add_words([attr["token"]])

    @classmethod
    def parsing(cls, raw_text_data, nlp_processor, processor_args):
        '''

        Parameters
        ----------
        raw_text_data: str
        nlp_processor: StanfordCoreNLP
        processor_args: dict

        Returns
        -------
        parsed_results: list[dict]
            Each sentence is a dict. All sentences are packed by a list.
            key, value
            "node_num": int
                the node amount
            "node_content": list[dict]
                The list consisting node information. Each node is organized by a dict.
                'token': str
                    word token
                'position_id': int
                    the word's position id in original sentence. eg: I am a dog. position_id: 0, 1, 2, 3
                'id': int,
                    the node token's id which will be used in GraphData
                "sentence_id": int
                    The sentence's id in the whole text.
            "graph_content": list[dict]
                The list consisting edge information. Each edge is organized by a dict.
                "edge_type": str
                    The edge type token, eg: 'nsubj'
                'src': int
                    The source node ``id``
                'tgt': int
                    The target node ``id``
        '''
        dep_json = nlp_processor.annotate(raw_text_data.strip(), properties=processor_args)
        from .utils import CORENLP_TIMEOUT_SIGNATURE
        if CORENLP_TIMEOUT_SIGNATURE in dep_json:
            raise TimeoutError('CoreNLP timed out at input: \n{}\n This item will be skipped. '
                               'Please check the input or change the timeout threshold.'.format(raw_text_data))

        dep_dict = json.loads(dep_json)

        parsed_results = []
        node_id = 0
        for s_id in range(len(dep_dict["sentences"])):
            parsed_sent = []
            node_item = []
            unique_hash = {}
            node_id = 0

            for tokens in dep_dict["sentences"][s_id]['tokens']:
                unique_hash[(tokens['index'], tokens['word'])] = node_id
                node = {
                    'token': tokens["word"],
                    'position_id': tokens["index"] - 1,
                    'id': node_id,
                    "sentence_id": s_id
                }
                node_item.append(node)
                node_id += 1

            for dep in dep_dict["sentences"][s_id]["basicDependencies"]:
                if cls.verbase > 0:
                    print(dep)

                if dep['governorGloss'] == "ROOT":
                    continue

                if dep['dependentGloss'] == "ROOT":
                    continue

                dep_info = {
                    "edge_type": dep['dep'],
                    'src': unique_hash[(dep['governor'], dep['governorGloss'])],
                    'tgt': unique_hash[(dep['dependent'], dep['dependentGloss'])]
                }
                if cls.verbase > 0:
                    print(dep_info)
                parsed_sent.append(dep_info)
            if cls.verbase > 0:
                print(node_id)
                print(len(parsed_sent))
            parsed_results.append({
                "graph_content": parsed_sent,
                "node_content": node_item,
                "node_num": node_id
            })
        return parsed_results

    @classmethod
    def topology(cls, raw_text_data, nlp_processor, processor_args, merge_strategy, edge_strategy, sequential_link=True,
                 verbase=0):
        """
            Graph building method.

        Parameters
        ----------
        raw_text_data: str or list[list]
            Raw text data, it can be multi-sentences.
            When it is ``str`` type, it is the raw text.
            When it is ``list[list]`` type, it is the tokenized token lists.
        nlp_processor: StanfordCoreNLP
            NLP parsing tools
        processor_args: dict
            The configure dict for StanfordCoreNLP.annotate
        merge_strategy: None or str, option=[None, "tailhead", "user_define"]
            Strategy to merge sub-graphs into one graph
            ``None``: It will be the default option. We will do as ``"tailhead"``.
            ``"tailhead"``: Link the sub-graph  ``i``'s tail node with ``i+1``'s head node
            ``"user_define"``: We will give this option to the user. User can override this method to define your merge
                               strategy.
        edge_strategy: None or str, option=[None, "homogeneous", "heterogeneous", "as_node"]
            Strategy to process edge.
            ``None``: It will be the default option. We will do as ``"homogeneous"``.
            ``"homogeneous"``: We will drop the edge type information.
                               If there is a linkage among node ``i`` and node ``j``, we will add an edge whose weight
                               is ``1.0``. Otherwise there is no edge.
            ``heterogeneous``: We will keep the edge type information.
                               An edge will have type information like ``n_subj``.
            ``as_node``: We will view the edge as a graph node.
                         If there is an edge whose type is ``k`` between node ``i`` and node ``j``,
                         we will insert a node ``k`` into the graph and link node (``i``, ``k``) and (``k``, ``j``).

        sequential_link: bool, default=True
            Whether to link node tokens sequentially (note that it is bidirectional)
        verbase: int, default=0
            Whether to output log infors. Set 1 to output more infos.
        Returns
        -------
        joint_graph: GraphData
            The merged graph data-structure.
        """
        cls.verbase = verbase
            

        parsed_results = cls.parsing(raw_text_data=raw_text_data, nlp_processor=nlp_processor,
                                     processor_args=processor_args)

        sub_graphs = []
        for sent_id, parsed_sent in enumerate(parsed_results):
            graph = cls._construct_static_graph(parsed_sent, edge_strategy=edge_strategy,
                                                sequential_link=sequential_link)
            sub_graphs.append(graph)
        joint_graph = cls._graph_connect(sub_graphs, merge_strategy)
        return joint_graph

    @classmethod
    def _construct_static_graph(cls, parsed_object, edge_strategy=None, sequential_link=True):
        """
            Build dependency-parsing-tree based graph for single sentence.

        Parameters
        ----------
        parsed_object: dict
            The parsing tree.
        edge_strategy: None or str, option=[None, "homogeneous", "heterogeneous", "as_node"]
            Strategy to process edge.
            ``None``: It will be the default option. We will do as ``"homogeneous"``.
            ``"homogeneous"``: We will drop the edge type information.
                               If there is a linkage among node ``i`` and node ``j``, we will add an edge whose weight
                               is ``1.0``. Otherwise there is no edge.
            ``heterogeneous``: We will keep the edge type information.
                               An edge will have type information like ``n_subj``.
                               It is not implemented yet.
            ``as_node``: We will view the edge as a graph node.
                         If there is an edge whose type is ``k`` between node ``i`` and node ``j``,
                         we will insert a node ``k`` into the graph and link node (``i``, ``k``) and (``k``, ``j``).
                         It is not implemented yet.

        Returns
        -------
        graph: GraphData
            graph structure for single sentence
        """
        ret_graph = GraphData()
        node_num = parsed_object["node_num"]
        assert node_num > 0
        ret_graph.add_nodes(node_num)
        head_node = 0
        tail_node = node_num - 1

        # insert node attributes
        node_objects = parsed_object["node_content"]
        for node in node_objects:
            ret_graph.node_attributes[node['id']]['type'] = 0
            ret_graph.node_attributes[node['id']]['token'] = node['token']
            ret_graph.node_attributes[node['id']]['position_id'] = node['position_id']
            ret_graph.node_attributes[node['id']]['sentence_id'] = node['sentence_id']
            ret_graph.node_attributes[node['id']]['head'] = False
            ret_graph.node_attributes[node['id']]['tail'] = False

        for dep_info in parsed_object["graph_content"]:
            if edge_strategy is None or edge_strategy == "homogeneous":
                ret_graph.add_edge(dep_info["src"], dep_info['tgt'])
            elif edge_strategy == "heterogeneous":
                ret_graph.add_edge(dep_info["src"], dep_info['tgt'])
                edge_idx = ret_graph.edge_ids(dep_info["src"], dep_info['tgt'])[0]
                ret_graph.edge_attributes[edge_idx]["token"] = dep_info["edge_type"]
            elif edge_strategy == "as_node":
                # insert a node
                node_idx = ret_graph.get_node_num()
                ret_graph.add_nodes(1)
                ret_graph.node_attributes[node_idx]['type'] = 3  # 3 for edge node
                ret_graph.node_attributes[node_idx]['token'] = dep_info['edge_type']
                ret_graph.node_attributes[node_idx]['position_id'] = None
                ret_graph.node_attributes[node_idx]['head'] = False
                ret_graph.node_attributes[node_idx]['tail'] = False
                # add edge infos
                ret_graph.add_edge(dep_info['src'], node_idx)
                ret_graph.add_edge(node_idx, dep_info['tgt'])
            else:
                raise NotImplementedError()
        ret_graph.node_attributes[head_node]['head'] = True
        ret_graph.node_attributes[tail_node]['tail'] = True

        sequential_list = [i for i in range(node_num)]

        if sequential_list and len(sequential_list) > 1:
            for st, ed in zip(sequential_list[:-1], sequential_list[1:]):
                try:
                    ret_graph.edge_ids(st, ed)
                except:
                    ret_graph.add_edge(st, ed)
                try:
                    ret_graph.edge_ids(ed, st)
                except:
                    ret_graph.add_edge(ed, st)
        return ret_graph

    @classmethod
    def _graph_connect(cls, nx_graph_list, merge_strategy=None):
        """
            This method will merge the sub-graphs into one graph.

        Parameters
        ----------
        nx_graph_list: list[GraphData]
            The list of all sub-graphs.
        merge_strategy: None or str, option=[None, "tailhead", "user_define"]
            Strategy to merge sub-graphs into one graph
            ``None``: It will be the default option. We will do as ``"tailhead"``.
            ``"tailhead"``: Link the sub-graph  ``i``'s tail node with ``i+1``'s head node
            ``"user_define"``: We will give this option to the user. User can override this method to define your merge
                               strategy.

        Returns
        -------
        joint_graph: GraphData
            The merged graph structure.
        """
        if cls.verbase > 0:
            print("sub_graph print")
            for i, s_g in enumerate(nx_graph_list):
                print("-------------------------")
                print("sub-graph: {}".format(i))
                print("node_num: {}".format(s_g.get_node_num()))
                for i in range(s_g.get_node_num()):
                    print(s_g.get_node_attrs(i))
                print("edge_num: {}".format(s_g.get_edge_num()))
                print(s_g.get_all_edges())
                for i in range(s_g.get_edge_num()):
                    print(i, s_g.edge_attributes[i])
            print("*************************************")
        if len(nx_graph_list) == 1:
            return nx_graph_list[0]
        elif len(nx_graph_list) == 0:
            raise RuntimeError("There is no graph needed to merge.")
        node_num_list = [s_g.get_node_num() for s_g in nx_graph_list]
        node_num = sum(node_num_list)
        g = GraphData()
        g.add_nodes(node_num)
        node_idx_off = 0

        # copy edges
        for s_g in nx_graph_list:
            for edge in s_g.get_all_edges():
                src, tgt = edge
                edge_idx_old = s_g.edge_ids(src, tgt)[0]
                g.add_edge(src + node_idx_off, tgt + node_idx_off)
                edge_idx_new = g.edge_ids(src + node_idx_off, tgt + node_idx_off)[0]
                if cls.verbase > 0:
                    print(edge_idx_new, edge_idx_old)
                    print(s_g.edge_attributes[edge_idx_old], "--------")
                g.edge_attributes[edge_idx_new] = copy.deepcopy(s_g.edge_attributes[edge_idx_old])
            # tmp = {}
            for key, value in enumerate(s_g.node_attributes):
                g.node_attributes[key + node_idx_off] = copy.deepcopy(value)
            #     tmp[key + node_idx_off] = copy.deepcopy(value)
            # g.node_attributes.update(tmp)
            node_idx_off += s_g.get_node_num()

        headtail_list = []

        head = -1
        tail = -1

        node_idx_off = 0
        for i in range(len(nx_graph_list)):
            for node_idx, node_attrs in enumerate(nx_graph_list[i].node_attributes):
                if node_attrs['head'] is True:
                    head = node_idx + node_idx_off
                if node_attrs['tail'] is True:
                    tail = node_idx + node_idx_off
            assert head != -1
            assert tail != -1
            headtail_list.append((head, tail))
            head = -1
            tail = -1
            node_idx_off += node_num_list[i]

        head_g = headtail_list[0][0]
        tail_g = headtail_list[-1][1]

        if merge_strategy is None or merge_strategy == "tailhead":

            src_list = []
            tgt_list = []

            for i in range(len(headtail_list) - 1):
                src_list.append(headtail_list[i][1])
                tgt_list.append(headtail_list[i + 1][0])
            if cls.verbase > 0:
                print("merged edges")
                print("src list:", src_list)
                print("tgt list:", tgt_list)
            g.add_edges(src_list, tgt_list)
        else:
            raise NotImplementedError()

        for node_idx, node_attrs in enumerate(g.node_attributes):
            node_attrs['head'] = node_idx == head_g
            node_attrs['tail'] = node_idx == tail_g

        if cls.verbase > 0:
            print("-----------------------------")
            print("merged graph")
            print("node_num: {}".format(g.get_node_num()))
            for i in range(g.get_node_num()):
                print(g.get_node_attrs(i))
            print("edge_num: {}".format(g.get_edge_num()))
            print(g.get_all_edges())
            for i in range(g.get_edge_num()):
                print(i, g.edge_attributes[i])
        return g

    def forward(self, batch_graphdata: list):
        batch_graphdata = self.embedding_layer(batch_graphdata)
        return batch_graphdata



from dgl.nn.pytorch import GraphConv

import torch
import torch.nn as nn
import dgl.function as fn
from torch.nn import init
from .base import GNNLayerBase, GNNBase
from ...data.data import GraphData
from dgl.utils import expand_as_pair

class GCN(GNNBase):
    r"""Multi-layer Graph Convolutional Networks (GCN).
    Support both `Unidirectional GCN
    <https://arxiv.org/pdf/1609.02907>`__ and bidirectional versions
    including `GCN-BiSep` and `GCN-BiFuse`.

    For the Unidirectional GCN,

    .. math::
      h_i^{(l+1)} = \sigma(b^{(l)} + \sum_{j\in\mathcal{N}(i)}\frac{1}{c_{ij}}h_j^{(l)}W^{(l)})

    where :math:`\mathcal{N}(i)` is the set of neighbors of node :math:`i`,
    :math:`c_{ij}` is the product of the square root of node degrees
    (i.e.,  :math:`c_{ij} = \sqrt{|\mathcal{N}(i)|}\sqrt{|\mathcal{N}(j)|}`),
    and :math:`\sigma` is an activation function.

    Parameters
    ----------
    num_layers: int
        Number of GCN layers.

    input_size: int
        Input feature size of the first GCN layer.

    hidden_size: int
        Hidden size per GCN layer.

    output_size: int
        Output feature size of the final GCN layer.

    direction_option: str
        Whether to use unidirectional (i.e., regular) or bidirectional (i.e., "bi_sep" and "bi_fuse") versions.
        Default : ``'bi_sep'``.

    gcn_norm: str, optional
        How to apply the normalizer. If is `'right'`, divide the aggregated messages
        by each node's in-degrees, which is equivalent to averaging the received messages.
        If is `'none'`, no normalization is applied. Default is `'both'`,
        where the :math:`c_{ij}` in the paper is applied.

    weight: bool, optional
        If True, apply a linear layer. Otherwise, aggregating the messages
        without a weight matrix.

    bias: bool, optional
        If True, adds a learnable bias to the output. Default: ``True``.

    activation: callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.

    allow_zero_in_degree: bool, optional

    use_edge_weight: bool, optional
    """
    def __init__(self,
                 num_layers,
                 input_size,
                 hidden_size,
                 output_size,
                 direction_option='bi_sep',
                 feat_drop=0.,
                 gcn_norm='both',
                 weight=True,
                 bias=True,
                 activation=None,
                 allow_zero_in_degree=False,
                 use_edge_weight=False,
                 residual=True):
        super(GCN, self).__init__()
        self.num_layers = num_layers
        self.direction_option = direction_option
        self.gcn_layers = nn.ModuleList()
        assert self.num_layers > 0
        self.use_edge_weight = use_edge_weight

        if isinstance(hidden_size, int):
            hidden_size = [hidden_size] * (self.num_layers - 1)

        if self.num_layers > 1:
            # input projection
            self.gcn_layers.append(GCNLayer(input_size,
                                            hidden_size[0],
                                            direction_option=self.direction_option,
                                            feat_drop=feat_drop,
                                            gcn_norm=gcn_norm,
                                            weight=weight,
                                            bias=bias,
                                            activation=activation,
                                            allow_zero_in_degree=allow_zero_in_degree,
                                            residual=residual))

        # hidden layers
        for l in range(1, self.num_layers - 1):
            # due to multi-head, the input_size = hidden_size * num_heads
            self.gcn_layers.append(GCNLayer(hidden_size[l - 1],
                                            hidden_size[l],
                                            direction_option=self.direction_option,
                                            feat_drop=feat_drop,
                                            gcn_norm=gcn_norm,
                                            weight=weight,
                                            bias=bias,
                                            activation=activation,
                                            allow_zero_in_degree=allow_zero_in_degree,
                                            residual=residual))
        # output projection
        self.gcn_layers.append(GCNLayer(hidden_size[-1] if self.num_layers > 1 else input_size,
                                        output_size,
                                        direction_option=self.direction_option,
                                        feat_drop=feat_drop,
                                        gcn_norm=gcn_norm,
                                        weight=weight,
                                        bias=bias,
                                        activation=activation,
                                        allow_zero_in_degree=allow_zero_in_degree,
                                        residual=residual))

    def forward(self, graph):
        r"""Compute multi-layer graph convolutional networks.

        Parameters
        ----------
        graph : GraphData
            The graph data containing topology and features.

        Returns
        -------
        GraphData
            The output graph data containing updated embeddings.
        """
        feat = graph.node_features['node_feat']
        dgl_graph = graph.to_dgl()

        if self.direction_option == 'bi_sep':
            h = [feat, feat]
        else:
            h = feat

        if self.use_edge_weight:
            edge_weight = graph.edge_features['edge_weight']
            if self.direction_option != 'undirected':
                reverse_edge_weight = graph.edge_features['reverse_edge_weight']
            else:
                reverse_edge_weight = None
        else:
            edge_weight = None
            reverse_edge_weight = None

        for l in range(self.num_layers - 1):
            h = self.gcn_layers[l](dgl_graph, h, edge_weight=edge_weight, reverse_edge_weight=reverse_edge_weight)
            if self.direction_option == 'bi_sep':
                h = [each.flatten(1) for each in h]
            else:
                h = h.flatten(1)

        # output projection
        logits = self.gcn_layers[-1](dgl_graph, h)

        if self.direction_option == 'bi_sep':
            # logits = [each.mean(1) for each in logits]
            logits = torch.cat(logits, -1)
        else:
            pass

        graph.node_features['node_emb'] = logits

        return graph

class GCNLayer(GNNLayerBase):
    r"""Single-layer GCN.

    Parameters
    ----------
    num_layers: int
        Number of GCN layers.

    input_size: int
        Input feature size of the first GCN layer.

    hidden_size: int
        Hidden size per GCN layer.

    output_size: int
        Output feature size of the final GCN layer.

    direction_option: str
        Whether to use unidirectional (i.e., regular) or bidirectional (i.e., "bi_sep" and "bi_fuse") versions.
        Default : ``'bi_sep'``.

    gcn_norm: str, optional
        How to apply the normalizer. If is `'right'`, divide the aggregated messages
        by each node's in-degrees, which is equivalent to averaging the received messages.
        If is `'none'`, no normalization is applied. Default is `'both'`,
        where the :math:`c_{ij}` in the paper is applied.

    weight: bool, optional
        If True, apply a linear layer. Otherwise, aggregating the messages
        without a weight matrix.

    bias: bool, optional
        If True, adds a learnable bias to the output. Default: ``True``.

    activation: callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.

    allow_zero_in_degree : bool, optional
        If there are 0-in-degree nodes in the graph, output for those nodes will be invalid
        since no message will be passed to those nodes. This is harmful for some applications
        causing silent performance regression. This module will raise a DGLError if it detects
        0-in-degree nodes in input graph. By setting ``True``, it will suppress the check
        and let the users handle it by themselves. Default: ``False``.

    """
    def __init__(self,
                 input_size,
                 output_size,
                 direction_option='bi_sep',
                 feat_drop=0.,
                 gcn_norm='both',
                 weight=True,
                 bias=True,
                 activation=None,
                 allow_zero_in_degree=False,
                 residual=True):
        super(GCNLayer, self).__init__()
        if direction_option == 'undirected':
            self.model = UndirectedGCNLayerConv( input_size,
                                                 output_size,
                                                 feat_drop=feat_drop,
                                                 gcn_norm=gcn_norm,
                                                 weight=weight,
                                                 bias=bias,
                                                 activation=activation,
                                                 allow_zero_in_degree=allow_zero_in_degree,
                                                 residual=residual)
        elif direction_option == 'bi_sep':
            self.model = BiSepGCNLayerConv(  input_size,
                                             output_size,
                                             feat_drop=feat_drop,
                                             gcn_norm=gcn_norm,
                                             weight=weight,
                                             bias=bias,
                                             activation=activation,
                                             allow_zero_in_degree=allow_zero_in_degree,
                                             residual=residual)
        elif direction_option == 'bi_fuse':
            self.model = BiFuseGCNLayerConv( input_size,
                                             output_size,
                                             feat_drop=feat_drop,
                                             gcn_norm=gcn_norm,
                                             weight=weight,
                                             bias=bias,
                                             activation=activation,
                                             allow_zero_in_degree=allow_zero_in_degree,
                                             residual=residual)
        else:
            raise RuntimeError('Unknown `direction_option` value: {}'.format(direction_option))

    def forward(self, graph, feat, weight=None, edge_weight=None, reverse_edge_weight=None):
        r"""Compute graph convolutional network layer.

        Parameters
        ----------
        graph : DGLGraph
            The graph.

        feat : torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.

        weight: torch.Tensor, optional
            Optional external weight tensor.

        edge_weight: torch.Tensor
            Optional edge weight. edge_weight shape: "math:`(\text{num_edge}, 1)`.

        reverse_edge_weight: torch.Tensor
            Optional reverse edge weight. reverse_edge_weight shape: "math:`(\text{num_edge}, 1)`.
        """
        return self.model(graph, feat, weight, edge_weight, reverse_edge_weight)


class UndirectedGCNLayerConv(GNNLayerBase):
    r"""

    Description
    -----------
    Graph convolution was introduced in `GCN <https://arxiv.org/abs/1609.02907>`__
    and mathematically is defined as follows:

    .. math::
      h_i^{(l+1)} = \sigma(b^{(l)} + \sum_{j\in\mathcal{N}(i)}\frac{1}{c_{ij}}h_j^{(l)}W^{(l)})

    where :math:`\mathcal{N}(i)` is the set of neighbors of node :math:`i`,
    :math:`c_{ij}` is the product of the square root of node degrees
    (i.e.,  :math:`c_{ij} = \sqrt{|\mathcal{N}(i)|}\sqrt{|\mathcal{N}(j)|}`),
    and :math:`\sigma` is an activation function.

    Parameters
    ----------
    input_size : int
        Input feature size; i.e, the number of dimensions of :math:`h_j^{(l)}`.
    output_size : int
        Output feature size; i.e., the number of dimensions of :math:`h_i^{(l+1)}`.
    gcn_norm : str, optional
        How to apply the normalizer. If is `'right'`, divide the aggregated messages
        by each node's in-degrees, which is equivalent to averaging the received messages.
        If is `'none'`, no normalization is applied. Default is `'both'`,
        where the :math:`c_{ij}` in the paper is applied.
    weight : bool, optional
        If True, apply a linear layer. Otherwise, aggregating the messages
        without a weight matrix.
    bias : bool, optional
        If True, adds a learnable bias to the output. Default: ``True``.
    activation : callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    allow_zero_in_degree : bool, optional
        If there are 0-in-degree nodes in the graph, output for those nodes will be invalid
        since no message will be passed to those nodes. This is harmful for some applications
        causing silent performance regression. This module will raise a DGLError if it detects
        0-in-degree nodes in input graph. By setting ``True``, it will suppress the check
        and let the users handle it by themselves. Default: ``False``.

    Attributes
    ----------
    weight : torch.Tensor
        The learnable weight tensor.
    bias : torch.Tensor
        The learnable bias tensor.
    """

    def __init__(self,
                 input_size,
                 output_size,
                 feat_drop=0.,
                 gcn_norm='both',
                 weight=True,
                 bias=True,
                 activation=None,
                 allow_zero_in_degree=False,
                 residual=True):
        super(UndirectedGCNLayerConv, self).__init__()
        if gcn_norm not in ('none', 'both', 'right'):
            raise RuntimeError('Invalid gcn_norm value. Must be either "none", "both" or "right".'
                               ' But got "{}".'.format(gcn_norm))
        self._input_size = input_size
        self._output_size = output_size
        self._gcn_norm = gcn_norm
        self._allow_zero_in_degree = allow_zero_in_degree
        self._feat_drop = nn.Dropout(feat_drop)

        if weight:
            self.weight = nn.Parameter(torch.Tensor(input_size, output_size))
        else:
            self.register_parameter('weight', None)

        if bias:
            self.bias = nn.Parameter(torch.Tensor(output_size))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

        if residual:
            if self._input_size != output_size:
                self.res_fc = nn.Linear(
                    self._input_size, output_size, bias=True)
            else:
                self.res_fc = nn.Identity()
        else:
            self.register_buffer('res_fc', None)

        self._activation = activation

    def reset_parameters(self):
        r"""
    
        Description
        -----------
        Reinitialize learnable parameters.
    
        Note
        ----
        The model parameters are initialized as in the
        `original implementation <https://github.com/tkipf/gcn/blob/master/gcn/layers.py>`__
        where the weight :math:`W^{(l)}` is initialized using Glorot uniform initialization
        and the bias is initialized to be zero.
    
        """
        if self.weight is not None:
            init.xavier_uniform_(self.weight)
        if self.bias is not None:
            init.zeros_(self.bias)
    
    def set_allow_zero_in_degree(self, set_value):
        r"""
    
        Description
        -----------
        Set allow_zero_in_degree flag.
    
        Parameters
        ----------
        set_value : bool
            The value to be set to the flag.
        """
        self._allow_zero_in_degree = set_value

    def forward(self, graph, feat, weight=None, edge_weight=None, reverse_edge_weight=None):
        r"""Compute graph convolution.

        Parameters
        ----------
        graph: DGLGraph
            The graph.

        feat: torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.

        weight: torch.Tensor, optional
            Optional external weight tensor.

        edge_weight: torch.Tensor
            Optional edge weight. edge_weight shape: "math:`(\text{num_edge}, 1)`.

        reverse_edge_weight: torch.Tensor
            Optional reverse edge weight. reverse_edge_weight shape: "math:`(\text{num_edge}, 1)`.
            For undirected GCN layer, reverse_edge_weight must be `None`.

        Returns
        -------
        torch.Tensor
            The output feature
        """
        assert reverse_edge_weight is None
        graph = graph.local_var()

        feat_origin = feat
        feat = self._feat_drop(feat)

        if self._gcn_norm == 'both':
            degs = graph.out_degrees().to(feat.device).float().clamp(min=1)
            gcn_norm = torch.pow(degs, -0.5)
            shp = gcn_norm.shape + (1,) * (feat.dim() - 1)
            gcn_norm = torch.reshape(gcn_norm, shp)
            feat = feat * gcn_norm

        if weight is not None:
            if self.weight is not None:
                raise RuntimeError('External weight is provided while at the same time the'
                                   ' module has defined its own weight parameter. Please'
                                   ' create the module with flag weight=False.')
        else:
            weight = self.weight

        if self._input_size > self._output_size:
            # mult W first to reduce the feature size for aggregation.
            if weight is not None:
                feat = torch.matmul(feat, weight)
            graph.srcdata['h'] = feat
            if edge_weight is None:
                graph.update_all(fn.copy_src(src='h', out='m'),
                                 fn.sum(msg='m', out='h'))
            else:
                graph.edata['edge_weight'] = edge_weight
                graph.update_all(fn.u_mul_e('h', 'edge_weight', 'm'),
                                 fn.sum('m', 'h'))
            rst = graph.dstdata['h']
        else:
            # aggregate first then mult W
            graph.srcdata['h'] = feat
            if edge_weight is None:
                graph.update_all(fn.copy_src(src='h', out='m'),
                                 fn.sum(msg='m', out='h'))
            else:
                graph.edata['edge_weight'] = edge_weight
                graph.update_all(fn.u_mul_e('h', 'edge_weight', 'm'),
                                 fn.sum('m', 'h'))
            rst = graph.dstdata['h']
            if weight is not None:
                rst = torch.matmul(rst, weight)

        if self._gcn_norm != 'none':
            degs = graph.in_degrees().to(feat.device).float().clamp(min=1)
            if self._gcn_norm == 'both':
                gcn_norm = torch.pow(degs, -0.5)
            else:
                gcn_norm = 1.0 / degs
            shp = gcn_norm.shape + (1,) * (feat.dim() - 1)
            gcn_norm = torch.reshape(gcn_norm, shp)
            rst = rst * gcn_norm

        if self.bias is not None:
            rst = rst + self.bias

        if self.res_fc is not None:
            h_dst = feat_origin
            resval = self.res_fc(h_dst).view(h_dst.shape[0], self._output_size)
            rst = rst + resval

        if self._activation is not None:
            rst = self._activation(rst)

        return rst


    def extra_repr(self):
        """Set the extra representation of the module,
        which will come into effect when printing the model.
        """
        summary = 'in={_input_size}, out={_output_size}'
        summary += ', gcn_normalization={_gcn_norm}'
        if '_activation' in self.__dict__:
            summary += ', activation={_activation}'
        return summary.format(**self.__dict__)


class BiFuseGCNLayerConv(GNNLayerBase):
    r"""Bidirection version GCN layer from paper `GCN <https://arxiv.org/abs/1609.02907>`__.

    .. math::
        h_{i, \vdash}^{(l+1)} = \sigma(b^{(l)}_{\vdash} + \sum_{j\in\mathcal{N}_{\vdash}(i)}\frac{1}{c_{ij}}h_{j}^{(l)}W^{(l)}_{\vdash})

        h_{i, \dashv}^{(l+1)} = \sigma(b^{(l)}_{\dashv} + \sum_{j\in\mathcal{N}_{\dashv}(i)}\frac{1}{c_{ij}}h_{j}^{(l)}W^{(l)}_{\dashv})

        r_{i}^{l} &= \sigma (W_{f}[h_{i, \vdash}^{l};h_{i, \dashv}^{l};
                h_{i, \vdash}^{l}*h_{i, \dashv}^{l};
                h_{i, \vdash}^{l}-h_{i, \dashv}^{l}])

    Parameters
    ----------
    input_size : int
        Input feature size; i.e, the number of dimensions of :math:`h_j^{(l)}`.
    output_size : int
        Output feature size; i.e., the number of dimensions of :math:`h_i^{(l+1)}`.
    gcn_norm : str, optional
        How to apply the normalizer. If is `'right'`, divide the aggregated messages
        by each node's in-degrees, which is equivalent to averaging the received messages.
        If is `'none'`, no normalization is applied. Default is `'both'`,
        where the :math:`c_{ij}` in the paper is applied.
    weight : bool, optional
        If True, apply a linear layer. Otherwise, aggregating the messages
        without a weight matrix.
    bias : bool, optional
        If True, adds a learnable bias to the output. Default: ``True``.
    activation : callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    allow_zero_in_degree : bool, optional
        If there are 0-in-degree nodes in the graph, output for those nodes will be invalid
        since no message will be passed to those nodes. This is harmful for some applications
        causing silent performance regression. This module will raise a DGLError if it detects
        0-in-degree nodes in input graph. By setting ``True``, it will suppress the check
        and let the users handle it by themselves. Default: ``False``.
    """

    def __init__(self,
                 input_size,
                 output_size,
                 feat_drop=0.,
                 gcn_norm='both',
                 weight=True,
                 bias=True,
                 activation=None,
                 allow_zero_in_degree=False,
                 residual=True):
        super(BiFuseGCNLayerConv, self).__init__()
        if gcn_norm not in ('none', 'both', 'right'):
            raise RuntimeError('Invalid gcn_norm value. Must be either "none", "both" or "right".'
                               ' But got "{}".'.format(gcn_norm))
        self._input_size = input_size
        self._output_size = output_size
        self._gcn_norm = gcn_norm
        self._allow_zero_in_degree = allow_zero_in_degree
        self._feat_drop=nn.Dropout(feat_drop)

        if weight:
            self.weight_fw = nn.Parameter(torch.Tensor(input_size, output_size))
            self.weight_bw = nn.Parameter(torch.Tensor(input_size, output_size))
        else:
            self.register_parameter('weight_fw', None)
            self.register_parameter('weight_bw', None)

        if bias:
            self.bias_fw = nn.Parameter(torch.Tensor(output_size))
            self.bias_bw = nn.Parameter(torch.Tensor(output_size))
        else:
            self.register_parameter('bias_fw', None)
            self.register_parameter('bias_bw', None)

        self.reset_parameters()

        self._activation = activation

        self.fuse_linear = nn.Linear(4 * output_size, output_size, bias=True)

        if residual:
            if self._input_size != output_size:
                self.res_fc = nn.Linear(
                    self._input_size, output_size, bias=True)
            else:
                self.res_fc = nn.Identity()
        else:
            self.register_buffer('res_fc', None)


    def reset_parameters(self):
        r"""
        Reinitialize learnable parameters.
        """
        if self.weight_fw is not None:
            init.xavier_uniform_(self.weight_fw)
            init.xavier_uniform_(self.weight_bw)
        if self.bias_fw is not None:
            init.zeros_(self.bias_fw)
            init.zeros_(self.bias_bw)

    def set_allow_zero_in_degree(self, set_value):
        r"""

        Description
        -----------
        Set allow_zero_in_degree flag.

        Parameters
        ----------
        set_value : bool
            The value to be set to the flag.
        """
        self._allow_zero_in_degree = set_value

    def forward(self, graph, feat, weight=None, edge_weight=None, reverse_edge_weight=None):
        r"""

        Description
        -----------
        Compute graph convolution.

        Parameters
        ----------
        graph: DGLGraph
            The graph.

        feat: torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.

        weight: torch.Tensor, optional
            Optional external weight tensor.

        edge_weight: torch.Tensor
            Optional edge weight. edge_weight shape: "math:`(\text{num_edge}, 1)`.

        reverse_edge_weight: torch.Tensor
            Optional reverse edge weight. reverse_edge_weight shape: "math:`(\text{num_edge}, 1)`.

        Returns
        -------
        torch.Tensor
            The output feature
        """
        feat_fw = feat_bw = self._feat_drop(feat)
        if isinstance(weight, tuple):
            weight_fw, weight_bw = weight
        else:
            weight_fw = weight_bw = weight

        # forward direction
        with graph.local_scope():
            graph = graph.local_var()

            if self._gcn_norm == 'both':
                degs = graph.out_degrees().to(feat_fw.device).float().clamp(min=1)
                gcn_norm = torch.pow(degs, -0.5)
                shp = gcn_norm.shape + (1,) * (feat_fw.dim() - 1)
                gcn_norm = torch.reshape(gcn_norm, shp)
                feat_fw = feat_fw * gcn_norm

            if weight_fw is not None:
                if self.weight_fw is not None:
                    raise RuntimeError('External weight is provided while at the same time the'
                                       ' module has defined its own weight parameter. Please'
                                       ' create the module with flag weight=False.')
            else:
                weight_fw = self.weight_fw

            if self._input_size > self._output_size:
                # mult W first to reduce the feature size for aggregation.
                if weight_fw is not None:
                    feat_fw = torch.matmul(feat_fw, weight_fw)
                graph.srcdata['h'] = feat_fw
                if edge_weight is None:
                    graph.update_all(fn.copy_src(src='h', out='m'),
                                     fn.sum(msg='m', out='h'))
                else:
                    graph.edata['edge_weight'] = edge_weight
                    graph.update_all(fn.u_mul_e('h', 'edge_weight', 'm'),
                                     fn.sum('m', 'h'))
                rst_fw = graph.dstdata['h']
            else:
                # aggregate first then mult W
                graph.srcdata['h'] = feat_fw
                if edge_weight is None:
                    graph.update_all(fn.copy_src(src='h', out='m'),
                                     fn.sum(msg='m', out='h'))
                else:
                    graph.edata['edge_weight'] = edge_weight
                    graph.update_all(fn.u_mul_e('h', 'edge_weight', 'm'),
                                     fn.sum('m', 'h'))
                rst_fw = graph.dstdata['h']
                if weight_fw is not None:
                    rst_fw = torch.matmul(rst_fw, weight_fw)

            if self._gcn_norm != 'none':
                degs = graph.in_degrees().to(feat_fw.device).float().clamp(min=1)
                if self._gcn_norm == 'both':
                    gcn_norm = torch.pow(degs, -0.5)
                else:
                    gcn_norm = 1.0 / degs
                shp = gcn_norm.shape + (1,) * (feat_fw.dim() - 1)
                gcn_norm = torch.reshape(gcn_norm, shp)
                rst_fw = rst_fw * gcn_norm

            if self.bias_fw is not None:
                rst_fw = rst_fw + self.bias_fw

            if self._activation is not None:
                rst_fw = self._activation(rst_fw)

        # backward direction
        graph = graph.reverse()
        with graph.local_scope():
            graph = graph.local_var()

            if self._gcn_norm == 'both':
                degs = graph.out_degrees().to(feat_bw.device).float().clamp(min=1)
                gcn_norm = torch.pow(degs, -0.5)
                shp = gcn_norm.shape + (1,) * (feat_bw.dim() - 1)
                gcn_norm = torch.reshape(gcn_norm, shp)
                feat_bw = feat_bw * gcn_norm

            if weight_bw is not None:
                if self.weight_bw is not None:
                    raise RuntimeError('External weight is provided while at the same time the'
                                       ' module has defined its own weight parameter. Please'
                                       ' create the module with flag weight=False.')
            else:
                weight_bw = self.weight_bw

            if self._input_size > self._output_size:
                # mult W first to reduce the feature size for aggregation.
                if weight_bw is not None:
                    feat_bw = torch.matmul(feat_bw, weight_bw)
                graph.srcdata['h'] = feat_bw
                if reverse_edge_weight is None:
                    graph.update_all(fn.copy_src(src='h', out='m'),
                                     fn.sum(msg='m', out='h'))
                else:
                    graph.edata['reverse_edge_weight'] = reverse_edge_weight
                    graph.update_all(fn.u_mul_e('h', 'reverse_edge_weight', 'm'),
                                     fn.sum('m', 'h'))
                rst_bw = graph.dstdata['h']
            else:
                # aggregate first then mult W
                graph.srcdata['h'] = feat_bw
                if edge_weight is None:
                    graph.update_all(fn.copy_src(src='h', out='m'),
                                     fn.sum(msg='m', out='h'))
                else:
                    graph.edata['reverse_edge_weight'] = reverse_edge_weight
                    graph.update_all(fn.u_mul_e('h', 'reverse_edge_weight', 'm'),
                                     fn.sum('m', 'h'))
                rst_bw = graph.dstdata['h']
                if weight_bw is not None:
                    rst_bw = torch.matmul(rst_bw, weight_bw)

            if self._gcn_norm != 'none':
                degs = graph.in_degrees().to(feat_bw.device).float().clamp(min=1)
                if self._gcn_norm == 'both':
                    gcn_norm = torch.pow(degs, -0.5)
                else:
                    gcn_norm = 1.0 / degs
                shp = gcn_norm.shape + (1,) * (feat_bw.dim() - 1)
                gcn_norm = torch.reshape(gcn_norm, shp)
                rst_bw = rst_bw * gcn_norm

            if self.bias_bw is not None:
                rst_bw = rst_bw + self.bias_bw

            if self._activation is not None:
                rst_bw = self._activation(rst_bw)

        fuse_vector = torch.cat(
            [rst_fw, rst_bw, rst_fw * rst_bw, rst_fw - rst_bw], dim=-1)
        fuse_gate_vector = torch.sigmoid(self.fuse_linear(fuse_vector))
        rst = fuse_gate_vector * rst_fw + (1 - fuse_gate_vector) * rst_bw

        if self.res_fc is not None:
            h_dst = feat
            resval = self.res_fc(h_dst).view(h_dst.shape[0], self._output_size)
            rst = rst + resval

        if self._activation is not None:
            rst = self._activation(rst)

        return rst


class BiSepGCNLayerConv(GNNLayerBase):
    r"""Bidirection version GCN layer from paper `GCN <https://arxiv.org/abs/1609.02907>`__.

    .. math::
        h_{i, \vdash}^{(l+1)} = \sigma(b^{(l)}_{\vdash} + \sum_{j\in\mathcal{N}_{\vdash}(i)}\frac{1}{c_{ij}}h_{j, \vdash}^{(l)}W^{(l)}_{\vdash})

        h_{i, \dashv}^{(l+1)} = \sigma(b^{(l)}_{\dashv} + \sum_{j\in\mathcal{N}_{\dashv}(i)}\frac{1}{c_{ij}}h_{j, \dashv}^{(l)}W^{(l)}_{\dashv})
    """
    def __init__(self,
                 input_size,
                 output_size,
                 feat_drop=0.,
                 gcn_norm='both',
                 weight=True,
                 bias=True,
                 activation=None,
                 allow_zero_in_degree=False,
                 residual=True):
        super(BiSepGCNLayerConv, self).__init__()
        if gcn_norm not in ('none', 'both', 'right'):
            raise RuntimeError('Invalid gcn_norm value. Must be either "none", "both" or "right".'
                               ' But got "{}".'.format(gcn_norm))
        self._input_size = input_size
        self._output_size = output_size
        self._gcn_norm = gcn_norm
        self._allow_zero_in_degree = allow_zero_in_degree
        self._feat_drop=nn.Dropout(feat_drop)

        if weight:
            self.weight_fw = nn.Parameter(torch.Tensor(input_size, output_size))
            self.weight_bw = nn.Parameter(torch.Tensor(input_size, output_size))
        else:
            self.register_parameter('weight_fw', None)
            self.register_parameter('weight_bw', None)

        if bias:
            self.bias_fw = nn.Parameter(torch.Tensor(output_size))
            self.bias_bw = nn.Parameter(torch.Tensor(output_size))
        else:
            self.register_parameter('bias_fw', None)
            self.register_parameter('bias_bw', None)

        if residual:
            if self._input_size != output_size:
                self.res_fc_fw = nn.Linear(
                    self._input_size, output_size, bias=True)
                self.res_fc_bw = nn.Linear(
                    self._input_size, output_size, bias=True)
            else:
                self.res_fc_fw = self.res_fc_bw = nn.Identity()
        else:
            self.register_buffer('res_fc_fw', None)
            self.register_buffer('res_fc_bw', None)

        self.reset_parameters()

        self._activation = activation

    def reset_parameters(self):
        r"""
        Reinitialize learnable parameters.
        """
        if self.weight_fw is not None:
            init.xavier_uniform_(self.weight_fw)
            init.xavier_uniform_(self.weight_bw)
        if self.bias_fw is not None:
            init.zeros_(self.bias_fw)
            init.zeros_(self.bias_bw)

        if isinstance(self.res_fc_fw, nn.Linear):
            nn.init.xavier_normal_(self.res_fc_fw.weight)

        if isinstance(self.res_fc_bw, nn.Linear):
            nn.init.xavier_normal_(self.res_fc_bw.weight)

    def set_allow_zero_in_degree(self, set_value):
        r"""

        Description
        -----------
        Set allow_zero_in_degree flag.

        Parameters
        ----------
        set_value : bool
            The value to be set to the flag.
        """
        self._allow_zero_in_degree = set_value

    def forward(self, graph, feat, weight=None, edge_weight=None, reverse_edge_weight=None):
        r"""

        Description
        -----------
        Compute graph convolution.

        Parameters
        ----------
        graph: DGLGraph
            The graph.

        feat: torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.

        weight: torch.Tensor, optional
            Optional external weight tensor.

        edge_weight: torch.Tensor
            Optional edge weight. edge_weight shape: "math:`(\text{num_edge}, 1)`.

        reverse_edge_weight: torch.Tensor
            Optional reverse edge weight. reverse_edge_weight shape: "math:`(\text{num_edge}, 1)`.

        Returns
        -------
        torch.Tensor
            The output feature
        """
        feat_fw, feat_bw = feat
        feat_fw = self._feat_drop(feat_fw)
        feat_bw = self._feat_drop(feat_bw)
        if isinstance(weight, tuple):
            weight_fw, weight_bw = weight
        else:
            weight_fw = weight_bw = weight

        # forward direction
        with graph.local_scope():
            graph = graph.local_var()

            if self._gcn_norm == 'both':
                degs = graph.out_degrees().to(feat_fw.device).float().clamp(min=1)
                gcn_norm = torch.pow(degs, -0.5)
                shp = gcn_norm.shape + (1,) * (feat_fw.dim() - 1)
                gcn_norm = torch.reshape(gcn_norm, shp)
                feat_fw = feat_fw * gcn_norm

            if weight_fw is not None:
                if self.weight_fw is not None:
                    raise RuntimeError('External weight is provided while at the same time the'
                                       ' module has defined its own weight parameter. Please'
                                       ' create the module with flag weight=False.')
            else:
                weight_fw = self.weight_fw

            if self._input_size > self._output_size:
                # mult W first to reduce the feature size for aggregation.
                if weight_fw is not None:
                    feat_fw = torch.matmul(feat_fw, weight_fw)
                graph.srcdata['h'] = feat_fw
                if edge_weight is None:
                    graph.update_all(fn.copy_src(src='h', out='m'),
                                     fn.sum(msg='m', out='h'))
                else:
                    graph.edata['edge_weight'] = edge_weight
                    graph.update_all(fn.u_mul_e('h', 'edge_weight', 'm'),
                                     fn.sum('m', 'h'))
                rst_fw = graph.dstdata['h']
            else:
                # aggregate first then mult W
                graph.srcdata['h'] = feat_fw
                if edge_weight is None:
                    graph.update_all(fn.copy_src(src='h', out='m'),
                                     fn.sum(msg='m', out='h'))
                else:
                    graph.edata['edge_weight'] = edge_weight
                    graph.update_all(fn.u_mul_e('h', 'edge_weight', 'm'),
                                     fn.sum('m', 'h'))
                rst_fw = graph.dstdata['h']
                if weight_fw is not None:
                    rst_fw = torch.matmul(rst_fw, weight_fw)

            if self._gcn_norm != 'none':
                degs = graph.in_degrees().to(feat_fw.device).float().clamp(min=1)
                if self._gcn_norm == 'both':
                    gcn_norm = torch.pow(degs, -0.5)
                else:
                    gcn_norm = 1.0 / degs
                shp = gcn_norm.shape + (1,) * (feat_fw.dim() - 1)
                gcn_norm = torch.reshape(gcn_norm, shp)
                rst_fw = rst_fw * gcn_norm

            if self.bias_fw is not None:
                rst_fw = rst_fw + self.bias_fw

            # residual
            if self.res_fc_fw is not None:
                h_dst_fw = feat[0]
                resval_fw = self.res_fc_fw(h_dst_fw).view(h_dst_fw.shape[0], self._output_size)
                # .view(h_dst.shape[0], self._output_size)
                rst_fw = rst_fw + resval_fw

            if self._activation is not None:
                rst_fw = self._activation(rst_fw)

        # backward direction
        graph = graph.reverse()
        with graph.local_scope():
            graph = graph.local_var()

            if self._gcn_norm == 'both':
                degs = graph.out_degrees().to(feat_bw.device).float().clamp(min=1)
                gcn_norm = torch.pow(degs, -0.5)
                shp = gcn_norm.shape + (1,) * (feat_bw.dim() - 1)
                gcn_norm = torch.reshape(gcn_norm, shp)
                feat_bw = feat_bw * gcn_norm

            if weight_bw is not None:
                if self.weight_bw is not None:
                    raise RuntimeError('External weight is provided while at the same time the'
                                       ' module has defined its own weight parameter. Please'
                                       ' create the module with flag weight=False.')
            else:
                weight_bw = self.weight_bw

            if self._input_size > self._output_size:
                # mult W first to reduce the feature size for aggregation.
                if weight_bw is not None:
                    feat_bw = torch.matmul(feat_bw, weight_bw)
                graph.srcdata['h'] = feat_bw
                if reverse_edge_weight is None:
                    graph.update_all(fn.copy_src(src='h', out='m'),
                                     fn.sum(msg='m', out='h'))
                else:
                    graph.edata['reverse_edge_weight'] = reverse_edge_weight
                    graph.update_all(fn.u_mul_e('h', 'reverse_edge_weight', 'm'),
                                     fn.sum('m', 'h'))
                rst_bw = graph.dstdata['h']
            else:
                # aggregate first then mult W
                graph.srcdata['h'] = feat_bw
                if reverse_edge_weight is None:
                    graph.update_all(fn.copy_src(src='h', out='m'),
                                     fn.sum(msg='m', out='h'))
                else:
                    graph.edata['reverse_edge_weight'] = edge_weight
                    graph.update_all(fn.u_mul_e('h', 'reverse_edge_weight', 'm'),
                                     fn.sum('m', 'h'))
                rst_bw = graph.dstdata['h']
                if weight_bw is not None:
                    rst_bw = torch.matmul(rst_bw, weight_bw)

            if self._gcn_norm != 'none':
                degs = graph.in_degrees().to(feat_bw.device).float().clamp(min=1)
                if self._gcn_norm == 'both':
                    gcn_norm = torch.pow(degs, -0.5)
                else:
                    gcn_norm = 1.0 / degs
                shp = gcn_norm.shape + (1,) * (feat_bw.dim() - 1)
                gcn_norm = torch.reshape(gcn_norm, shp)
                rst_bw = rst_bw * gcn_norm

            if self.bias_bw is not None:
                rst_bw = rst_bw + self.bias_bw

            # residual
            if self.res_fc_bw is not None:
                h_dst_bw = feat[1]
                resval_bw = self.res_fc_bw(h_dst_bw).view(h_dst_bw.shape[0], self._output_size)
                rst_bw = rst_bw + resval_bw

            if self._activation is not None:
                rst_bw = self._activation(rst_bw)

        return [rst_fw, rst_bw]


class EvaluationMetricBase:
    """
        Base class for evaluation metric
    """

    def __init__(self):
        pass

    def calculate_scores(self, **kwargs):
        raise NotImplementedError()



import platform
import os

from setuptools import setup, find_packages
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

cuda_versions = {
    '1': '9.2',
    '2': '10.1',
    '3': '10.2',
    '4': '11.0',
    '5': '11.1',
    '6': 'cpu'
}


class MyBdistWheel(_bdist_wheel):
    def finalize_options(self):
        _bdist_wheel.finalize_options(self)
        self.root_is_pure = False


if __name__ == '__main__':
    os_tag = {
        'Windows': 'win_amd64',
        'Darwin': 'macosx_x86_64',
        'Linux': 'manylinux1_x86_64'
    }

    # Parse config
    try:
        with open('config', 'r') as f:
            version = f.readlines()[0].strip()
    except:
        raise FileNotFoundError('config file not found. Please run ./configure first.')

    cuda_version = cuda_versions[version]

    if cuda_version == '11.1':
        os.system("pip3 install torch==1.8.1+cu111 torchvision==0.9.1+cu111 torchaudio==0.8.1 -f https://download.pytorch.org/whl/torch_stable.html")

    if cuda_version == 'none':
        cuda_version = ''
    elif cuda_version.find('.') != -1:
        cuda_version = '-cu' + ''.join(cuda_version.split('.'))
    else:
        cuda_version = ''

    install_requirement = ['pythonds', 'nltk >= 3.5', 'stanfordcorenlp', 'scipy >= 1.5.2',
                           'scikit-learn >= 0.23.2', 'networkx >= 2.5', 'dgl{} >= 0.4'.format(cuda_version),
                           'ogb', 'torchtext', 'tqdm >= 4.29.0', 'pyyaml']
    # pytorch_requirement = 'torch >= 1.6.0' if platform.system() != 'Windows' else 'torch >= 1.8.1'
    # print("System: {}. PyTorch Requirement = {}".format(platform.system(), pytorch_requirement))
    # if cuda_version in ['-cu101', '-cu92']:
    #     pytorch_requirement += ', <=1.7.0'
    # install_requirement.append(pytorch_requirement)


    setup(
        name='graph4nlp{}'.format(cuda_version),
        version='0.2a10',
        description='A DGL and PyTorch based graph deep learning library for natural language processing',
        author='Graph4NLP Team',
        license='Apache 2.0',
        include_package_data=True,
        packages=find_packages('.', exclude=(
            "examples.*", "examples", "graph4nlp.pytorch.test.*", "graph4nlp.pytorch.test")),
        install_requires=install_requirement,
        platforms=os_tag[platform.system()]
    )
    print("Graph4NLP Python library installation finished. Please manually check Stanford CoreNLP"
          "(https://stanfordnlp.github.io/CoreNLP/) is installed and "
          "running in your environment.")



from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab
from .base import GeneralLossBase
import torch.nn as nn
import torch


class CrossEntropyLoss(GeneralLossBase):
    """
       The cross-entropy with ``mask`` support.
       Technically, it is a wrapper of ``nn.NLLLoss``.
    
    Parameters
    ----------
    vocab: Vocab
       The vocab for loss calculation.
    """
    def __init__(self, vocab: Vocab):
        super(CrossEntropyLoss, self).__init__()
        self.loss_func = nn.NLLLoss()
        self.VERY_SMALL_NUMBER = 1e-31
        self.vocab = vocab

    def forward(self, logits, label):
        """
            The loss calculation.
        
        Parameters
        ----------
        logits: torch.Tensor
            The probability with the shape of ``[batch_size, max_decoder_step, vocab_size]``. \
            Note that it is calculated by ``softmax``.

        label: torch.Tensor
            The ground-truth with the shape of ``[batch_size, max_decoder_step]``.

        Returns
        -------
        ce_loss: torch.Tensor
            The cross-entropy loss.
        """
        assert logits.shape[0:1] == label.shape[0:1]
        assert len(logits.shape) == 3
        log_prob = torch.log(logits + self.VERY_SMALL_NUMBER)
        batch_size = label.shape[0]
        step = label.shape[1]

        mask = 1 - label.data.eq(self.vocab.PAD).float()

        prob_select = torch.gather(log_prob.view(batch_size * step, -1), 1, label.view(-1, 1))

        prob_select_masked = - torch.masked_select(prob_select, mask.view(-1, 1).bool())
        loss = torch.mean(prob_select_masked)
        return loss



from operator import itemgetter
import pickle as pkl
import torch
from random import randint
import numpy as np
import copy
from graph4nlp.pytorch.data.data import to_batch
class Tree():
    def __init__(self):
        self.parent = None
        self.num_children = 0
        self.children = []

    def __str__(self, level=0):
        ret = ""
        for child in self.children:
            if isinstance(child, type(self)):
                ret += child.__str__(level+1)
            else:
                ret += "\t"*level + str(child) + "\n"
        return ret

    def add_child(self, c):
        if isinstance(c, type(self)):
            c.parent = self
        self.children.append(c)
        self.num_children = self.num_children + 1

    def to_string(self):
        r_list = []
        for i in range(self.num_children):
            if isinstance(self.children[i], Tree):
                r_list.append("( " + self.children[i].to_string() + " )")
            else:
                r_list.append(" "+str(self.children[i])+" ")
        return "".join(r_list)

    def to_text(self, form_manager):
        r_list = []
        for i in range(self.num_children):
            if isinstance(self.children[i], Tree):
                r_list.append("( " + self.children[i].to_text(form_manager) + " )")
            else:
                r_list.append(" "+form_manager.get_idx_symbol(self.children[i])+" ")
        return "".join(r_list)

    def to_list(self, tgt_vocab):
        r_list = []
        for i in range(self.num_children):
            if isinstance(self.children[i], type(self)):
                r_list.append(tgt_vocab.get_symbol_idx("("))
                cl = self.children[i].to_list(tgt_vocab)
                for k in range(len(cl)):
                    r_list.append(cl[k])
                r_list.append(tgt_vocab.get_symbol_idx(")"))
            else:
                r_list.append(self.children[i])
        return r_list

    @staticmethod
    def norm_tree(r_list, tgt_vocab):
        q = [Tree.convert_to_tree(r_list, 0, len(r_list), tgt_vocab)]
        head = 0
        while head < len(q):
            t = q[head]
            exchangeable_symbol_idx_list = tgt_vocab.get_symbol_idx_for_list(
                ['and', 'or'])
            if len(t.children) > 0 and (t.children[0] in exchangeable_symbol_idx_list):
                k = []
                for i in range(1, len(t.children)):
                    if isinstance(t.children[i], Tree):
                        k.append((t.children[i].to_string(), i))
                    else:
                        k.append((str(t.children[i]), i))
                sorted_t_dict = []
                k.sort(key=itemgetter(0))
                for key1 in k:
                    sorted_t_dict.append(t.children[key1[1]])

                for i in range(t.num_children-1):
                    t.children[i+1] = \
                        sorted_t_dict[i]
            for i in range(len(t.children)):
                if isinstance(t.children[i], Tree):
                    q.append(t.children[i])

            head = head + 1
        return q[0].to_list(tgt_vocab)

    @staticmethod
    def deduplicate_tree(r_list, tgt_vocab):
        q = [Tree.convert_to_tree(r_list, 0, len(r_list), tgt_vocab)]
        head = 0
        while head < len(q):
            t = q[head]
            if len(t.children) > 0:
                k = {}
                for i in range(len(t.children)):
                    if isinstance(t.children[i], Tree):
                        k[t.children[i].to_string().strip()] = i
                    else:
                        k[str(t.children[i]).strip()] = i
                cnt_deleted = 0
                for index in (list(range(len(t.children)))):
                    if index not in ([item[1] for item in k.items()]):
                        t.children.pop(index - cnt_deleted)
                        t.num_children = t.num_children - 1
                        cnt_deleted += 1

            for i in range(len(t.children)):
                if isinstance(t.children[i], Tree):
                    q.append(t.children[i])
            head = head + 1
        return q[0].to_list(tgt_vocab)

    @staticmethod
    def convert_to_tree(r_list, i_left, i_right, tgt_vocab):
        t = Tree()
        level = 0
        left = -1
        for i in range(i_left, i_right):
            if r_list[i] == tgt_vocab.get_symbol_idx('('):
                if level == 0:
                    left = i
                level = level + 1
            elif r_list[i] == tgt_vocab.get_symbol_idx(')'):
                level = level - 1
                if level == 0:
                    if i == left+1:
                        c = r_list[i]
                    else:
                        c = Tree.convert_to_tree(
                            r_list, left + 1, i, tgt_vocab)
                    t.add_child(c)
            elif level == 0:
                t.add_child(r_list[i])
        return t

class VocabForAll():
    def __init__(self, in_word_vocab, out_word_vocab, share_vocab):
        self.in_word_vocab = in_word_vocab
        self.out_word_vocab = out_word_vocab
        self.share_vocab = share_vocab

class Vocab():
    def __init__(self, lower_case=True, 
                       pretrained_word_emb_name=None, 
                       pretrained_word_emb_url=None, 
                       pretrained_word_emb_cache_dir=None,  
                       embedding_dims=300):
        self.symbol2idx = {}
        self.idx2symbol = {}
        self.vocab_size = 0

        self.pad_token = '<PAD>'
        self.start_token = '<START>'
        self.end_token = '<END>'
        self.unk_token = '<UNK>'
        self.non_terminal_token = '<NON-TERMINAL>'

        self.add_symbol(self.pad_token)
        self.add_symbol(self.start_token)
        self.add_symbol(self.end_token)
        self.add_symbol(self.unk_token)
        self.add_symbol(self.non_terminal_token)

        self.lower_case = lower_case
        self.embedding_dims = embedding_dims
        self.embeddings = None

        self.pretrained_word_emb_name = pretrained_word_emb_name
        self.pretrained_word_emb_url = pretrained_word_emb_url
        self.pretrained_word_emb_cache_dir = pretrained_word_emb_cache_dir

    def add_symbol(self, s):
        if s not in self.symbol2idx:
            self.symbol2idx[s] = self.vocab_size
            self.idx2symbol[self.vocab_size] = s
            self.vocab_size += 1
        return self.symbol2idx[s]

    def get_symbol_idx(self, s):
        if s not in self.symbol2idx:
            return self.get_symbol_idx(self.unk_token)
        else:
            return self.symbol2idx[s]

    def get_idx_symbol(self, idx):
        if idx not in self.idx2symbol:
            return self.unk_token
        return self.idx2symbol[idx]

    def init_from_file(self, fn, min_freq, max_vocab_size):
        print("loading vocabulary file: {}".format(fn))
        with open(fn, "r") as f:
            for line in f:
                l_list = line.strip().split('\t')
                c = int(l_list[1])
                if c >= min_freq:
                    self.add_symbol(l_list[0])
                if self.vocab_size > max_vocab_size:
                    break
        if self.pretrained_word_emb_name is None:
            self.randomize_embeddings(self.embedding_dims)
        else:
            # print("loadding pretrained embedding file in {}".format(self.pretrained_embedding_fn))
            self.load_embeddings(pretrained_word_emb_name=self.pretrained_word_emb_name, 
                                 pretrained_word_emb_url=self.pretrained_word_emb_url, 
                                 pretrained_word_emb_cache_dir=self.pretrained_word_emb_cache_dir)

    def init_from_list(self, arr, min_freq=1, max_vocab_size=100000):
        for word_, c_ in arr:
            if c_ >= min_freq:
                self.add_symbol(word_)
            if self.vocab_size > max_vocab_size:
                break
        if self.pretrained_word_emb_name is None:
            self.randomize_embeddings(self.embedding_dims)
        else:
            # print("loadding pretrained embedding file in {}".format(self.pretrained_embedding_fn))
            self.load_embeddings(pretrained_word_emb_name=self.pretrained_word_emb_name, 
                                 pretrained_word_emb_url=self.pretrained_word_emb_url, 
                                 pretrained_word_emb_cache_dir=self.pretrained_word_emb_cache_dir)

    def get_symbol_idx_for_list(self, l):
        r = []
        for i in range(len(l)):
            r.append(self.get_symbol_idx(l[i]))
        return r

    def get_idx_symbol_for_list(self, l):
        l = np.array(l)
        r = []
        for i in range(len(l)):
            r.append(self.get_idx_symbol(l[i]))
        return r

    def load_embeddings(self, pretrained_word_emb_name="840B", 
                              pretrained_word_emb_url=None, 
                              pretrained_word_emb_cache_dir=".vector_cache/", 
                              dtype=np.float32):
        """Load pretrained word embeddings for initialization"""
        from .vocab_utils import WordEmbModel
        word_model = WordEmbModel(pretrained_word_emb_name=pretrained_word_emb_name, 
                                pretrained_word_emb_url=pretrained_word_emb_url, 
                                pretrained_word_emb_cache_dir=pretrained_word_emb_cache_dir)

        word_list = list(self.symbol2idx.keys())

        word_emb, hit, cnt = word_model.get_vecs_by_tokens(tokens=word_list, lower_case_backup=self.lower_case)
        
        self.embeddings = word_emb.numpy()
        self.embeddings[self.get_symbol_idx(self.unk_token)] = np.zeros(word_model.dim)

    def randomize_embeddings(self, n_dims, scale=0.08):
        """Use random word embeddings for initialization."""
        vocab_size = self.vocab_size
        shape = (vocab_size, n_dims)
        self.embeddings = np.array(np.random.uniform(low=-scale, high=scale, size=shape), dtype=np.float32)
        self.embeddings[self.get_symbol_idx(self.pad_token)] = np.zeros(n_dims)

    def __getitem__(self, item):
        if type(item) is int:
            return self.get_idx_symbol(item)
        return self.get_symbol_idx(item)

    def __len__(self):
        return self.vocab_size

def to_cuda(x, device=None):
    if device:
        x = x.to(device)
    return x



import operator
from queue import PriorityQueue

import torch
import torch.nn as nn

from graph4nlp.pytorch.data.data import GraphData
from graph4nlp.pytorch.modules.prediction.generation.base import DecoderBase
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab
from copy import deepcopy


class StrategyBase(nn.Module):
    def __init__(self):
        super(StrategyBase, self).__init__()

    def generate(self, **kwargs):
        raise NotImplementedError()


class BeamSearchNode(object):
    def __init__(self, hiddenstate, enc_attn_weights_average, previousNode, wordId, logProb, length):
        self.h = hiddenstate
        self.enc_attn_weights_average = enc_attn_weights_average
        self.prevNode = previousNode
        self.wordid = wordId
        self.logp = logProb
        self.leng = length

    def eval(self, alpha=1.0):
        reward = 0
        # Add here a function for shaping a reward

        return self.logp / float(self.leng - 1 + 1e-6) + alpha * reward


class Hypothesis(object):
    def __init__(self, tokens, log_probs, dec_state, input_feed, num_non_words, enc_attn_weights, use_coverage, states_for_tree = None):
        self.tokens = tokens
        self.states_for_tree = states_for_tree # to record decoder hidden states to be sent to descendant decoding process.
        self.log_probs = log_probs
        self.dec_state = dec_state
        self.num_non_words = num_non_words
        self.enc_attn_weights = enc_attn_weights
        self.use_coverage = use_coverage
        self.input_feed = input_feed
        if not self.use_coverage:
            self.enc_attn_weights = []

    def __repr__(self):
        return repr(self.tokens)

    def __len__(self):
        return len(self.tokens) - self.num_non_words

    @property
    def avg_log_prob(self):
        return sum(self.log_probs) / len(self.log_probs)

    def create_next(self, token, log_prob, dec_state, input_feed, non_word, add_enc_attn_weights):
        if self.use_coverage:
            assert len(add_enc_attn_weights.shape) == 3
            assert add_enc_attn_weights.shape[0] == 1
            assert add_enc_attn_weights.shape[1] == 1
            enc_attn_weights_processed = [t.clone() for t in self.enc_attn_weights]
            enc_attn_weights_processed.append(add_enc_attn_weights)
        else:
            enc_attn_weights_processed = []
        return Hypothesis(tokens=self.tokens + [token], log_probs=self.log_probs + [log_prob],
                          dec_state=dec_state, input_feed=input_feed,
                          num_non_words=self.num_non_words + 1 if non_word else self.num_non_words,
                          enc_attn_weights=enc_attn_weights_processed, use_coverage=self.use_coverage,
                          states_for_tree=self.states_for_tree + [dec_state] if self.states_for_tree is not None else None)



class DecoderStrategy(StrategyBase):
    """
        The strategy for sequence decoding. Support beam seach only temporally.
    Parameters
    ----------
    beam_size: int
        The beam size for beam search.
    batch_graph: GraphData
        The input graph
    decoder: DecoderBase
        The decoder instance.
    rnn_type: str, option=["lstm", "gru"]
        The type of RNN. 
    use_copy: bool, default=False
        Whether use ``copy`` mechanism. See pointer network. Note that you must use attention first.
    use_coverage: bool, default=False
        Whether use ``coverage`` mechanism. Note that you must use attention first.
    max_decoder_step: int, default=50
        The maximal decoding step.

    """
    def __init__(self, beam_size, vocab, decoder: DecoderBase, rnn_type, use_copy=False, use_coverage=False,
                 max_decoder_step=50):
        super(DecoderStrategy, self).__init__()
        self.rnn_type = rnn_type
        self.beam_size = beam_size
        self.decoder = decoder
        self.vocab = vocab
        self.use_copy = use_copy
        self.use_coverage = use_coverage
        self.max_decoder_step = max_decoder_step

    def generate(self, batch_graph, oov_dict=None, topk=1):
        """
            Generate sequences using beam search.
        Parameters
        ----------
        batch_graph: GraphData
        oov_dict: VocabModel, default=None
            The vocabulary for copy mechanism.
        topk: int, default=1

        Returns
        -------
        prediction: list
        """
        params = self.decoder.extract_params(batch_graph)
        params['tgt_seq'] = None
        params['beam_size'] = self.beam_size
        params['topk'] = topk
        params['oov_dict'] = oov_dict
        return self._beam_search(**params)

    def _beam_search(self, graph_node_embedding, graph_node_mask=None, rnn_node_embedding=None,
                     graph_level_embedding=None,
                     graph_edge_embedding=None, graph_edge_mask=None, tgt_seq=None, src_seq=None, oov_dict=None,
                     beam_size=4, topk=1):

        assert 0 < topk <= beam_size
        len_in_words = False

        min_out_len = 1
        max_out_len = self.max_decoder_step + 1
        batch_size = graph_node_embedding.shape[0]

        decoder_state = self.decoder.get_decoder_init_state(rnn_type=self.rnn_type, batch_size=batch_size,
                                                            content=graph_level_embedding)
                                                     
        input_feed = torch.zeros(batch_size, self.decoder.input_feed_size).to(graph_node_embedding.device)


        batch_results = []
        for batch_idx in range(batch_size):
            
            if self.rnn_type == "lstm":
                single_decoder_state = (
                    decoder_state[0][:, batch_idx, :].unsqueeze(1), decoder_state[1][:, batch_idx, :].unsqueeze(1))
            elif self.rnn_type == "gru":
                single_decoder_state = decoder_state[:, batch_idx, :].unsqueeze(1)
            else:
                raise NotImplementedError("RNN Type {} is not implemented, expected in ``lstm`` or ``gru``"
                                          .format(self.rnn_type))

            single_graph_node_embedding = graph_node_embedding[batch_idx, :, :].unsqueeze(0).expand(beam_size, -1, -1).contiguous()
            
            single_graph_node_mask = graph_node_mask[batch_idx, :].unsqueeze(0).expand(beam_size, -1).contiguous() if graph_node_mask is not None else None
            single_rnn_node_embedding = rnn_node_embedding[batch_idx, :, :].unsqueeze(
                0).expand(beam_size, -1, -1).contiguous() if rnn_node_embedding is not None else None
            single_input_feed = input_feed[batch_idx, :].unsqueeze(0)

            step = 0
            results, backup_results = [], []

            hypos = [Hypothesis(tokens=[self.vocab.SOS], log_probs=[], dec_state=single_decoder_state, input_feed=single_input_feed, num_non_words=1, enc_attn_weights=[], use_coverage=self.use_coverage)]

            while len(hypos) > 0 and step <= self.max_decoder_step:
                n_hypos = len(hypos)
                if n_hypos < beam_size:
                    hypos.extend(deepcopy(hypos[-1]) for _ in range(beam_size - n_hypos)) # check deep copy

                decoder_input = torch.tensor([h.tokens[-1] for h in hypos]).to(graph_node_embedding.device)
                input_feed_beam = torch.cat([h.input_feed for h in hypos], dim=0)
                if self.rnn_type == "lstm":
                    single_decoder_state = (
                        torch.cat([h.dec_state[0] for h in hypos], 1), torch.cat([h.dec_state[1] for h in hypos], 1))
                elif self.rnn_type == "gru":
                    single_decoder_state = torch.cat([h.dec_state for h in hypos], 1)
                else:
                    raise NotImplementedError("RNN Type {} is not implemented, expected in ``lstm`` or ``gru``"
                                              .format(self.rnn_type))

                decoder_input = self.decoder._filter_oov(decoder_input)

                enc_tmp = []
                coverage_input = []
                if step > 0  and self.use_coverage:
                    for ii in range(step):
                        coverage_input.append(torch.cat([h.enc_attn_weights[ii] for h in hypos], dim=1))
                assert decoder_input.shape[0] == beam_size
                
                # define inputs
                decoder_input_pad = decoder_input
                input_feed_beam_pad = input_feed_beam
                single_decoder_state_pad = single_decoder_state
                single_graph_node_mask_pad = single_graph_node_mask
                single_graph_node_embedding_pad = single_graph_node_embedding
                single_rnn_node_embedding_pad = single_rnn_node_embedding


                decoder_output, input_feed_ret, single_decoder_state, dec_attn_scores, _ = \
                    self.decoder.decode_step(decoder_input=decoder_input_pad, input_feed=input_feed_beam_pad, rnn_state=single_decoder_state_pad,
                                             dec_input_mask=single_graph_node_mask_pad,
                                             encoder_out=single_graph_node_embedding_pad, rnn_emb=single_rnn_node_embedding_pad,
                                             enc_attn_weights_average=coverage_input,
                                             src_seq=src_seq[batch_idx, :].unsqueeze(0).expand(decoder_input_pad.shape[0], -1) if self.use_copy else None,
                                             oov_dict=oov_dict)
                decoder_output = decoder_output[:beam_size]
                input_feed_ret = input_feed_ret[:beam_size]
                single_decoder_state = (
                    single_decoder_state[0][:, :beam_size, :],
                    single_decoder_state[1][:, :beam_size, :]
                )
                dec_attn_scores = dec_attn_scores[:beam_size]
 
                decoder_output = torch.log(decoder_output + 1e-31)
                top_v, top_i = decoder_output.data.topk(beam_size)

                new_hypos = []
                for in_idx in range(n_hypos):
                    for out_idx in range(beam_size):
                        new_tok = top_i[in_idx][out_idx].item()
                        new_prob = top_v[in_idx][out_idx].item()
                        new_input_feed = input_feed_ret[in_idx, :].unsqueeze(0)
                        new_enc_attn_weights = dec_attn_scores[in_idx, :].unsqueeze(0).unsqueeze(0)

                        non_word = new_tok == self.vocab.EOS  # only SOS & EOS don't count

                        if self.rnn_type == "lstm":
                            tmp_decoder_state = (
                                single_decoder_state[0][:, in_idx, :].unsqueeze(1),
                                single_decoder_state[1][:, in_idx, :].unsqueeze(1))
                        elif self.rnn_type == "gru":
                            tmp_decoder_state = single_decoder_state[:, in_idx, :].unsqueeze(1)
                        else:
                            raise NotImplementedError("RNN Type {} is not implemented, expected in ``lstm`` or ``gru``"
                                                      .format(self.rnn_type))

                        new_hypo = hypos[in_idx].create_next(token=new_tok, log_prob=new_prob, input_feed=new_input_feed,
                                                             dec_state=tmp_decoder_state, non_word=non_word,
                                                             add_enc_attn_weights=new_enc_attn_weights)
                        new_hypos.append(new_hypo)

                # Block sequences with repeated ngrams
                # block_ngram_repeats(new_hypos, block_ngram_repeat) we will implement it in the future

                # process the new hypotheses
                new_hypos = sorted(new_hypos, key=lambda h: -h.avg_log_prob)[:beam_size]
                hypos = []
                new_complete_results, new_incomplete_results = [], []
                for nh in new_hypos:
                    length = len(nh)  # Does not count SOS and EOS
                    if nh.tokens[-1] == self.vocab.EOS:  # a complete hypothesis
                        if len(new_complete_results) < beam_size and min_out_len <= length <= max_out_len:
                            new_complete_results.append(nh)
                    elif len(hypos) < beam_size and length < max_out_len:  # an incomplete hypothesis
                        hypos.append(nh)
                    elif length == max_out_len and len(new_incomplete_results) < beam_size:
                        new_incomplete_results.append(nh)
                if new_complete_results:
                    results.extend(new_complete_results)
                elif new_incomplete_results:
                    backup_results.extend(new_incomplete_results)
                step += 1
            if not results:  # if no sequence ends with EOS within desired length, fallback to sequences
                results = backup_results  # that are "truncated" at the end to max_out_len
            batch_results.append(sorted(results, key=lambda h: -h.avg_log_prob)[:beam_size])
        ret = torch.zeros(batch_size, topk, self.max_decoder_step).long()
        for sent_id, each in enumerate(batch_results):
            for i in range(topk):
                ids = torch.Tensor(each[i].tokens[1:])[:self.max_decoder_step]
                if ids.shape[0] < self.max_decoder_step:
                    pad = torch.zeros(self.max_decoder_step - ids.shape[0])
                    ids = torch.cat((ids, pad), dim=0)
                ret[sent_id, i, :] = ids
        return ret

    def beam_search_for_tree_decoding(self, decoder_initial_state,
                                        decoder_initial_input,
                                        parent_state,
                                        graph_node_embedding,
                                        rnn_node_embedding=None,
                                        src_seq=None,
                                        oov_dict=None,
                                        sibling_state=None,
                                        device=None,
                                        topk=1,
                                        enc_batch=None):
        min_out_len = 1
        max_out_len = self.max_decoder_step
        batch_size = graph_node_embedding.size(0) 
        assert batch_size == 1
        decoder_hidden = decoder_initial_state

        batch_results = []
        for batch_idx in range(batch_size):
            single_graph_node_embedding = graph_node_embedding.expand(self.beam_size, -1, -1).contiguous()
            single_rnn_node_embedding = rnn_node_embedding.expand(self.beam_size, -1, -1).contiguous() if rnn_node_embedding is not None else None
            single_parent_state = parent_state.expand(self.beam_size, -1).contiguous() 

            step = 0
            results, backup_results = [], []
            hypos = [Hypothesis(tokens=[decoder_initial_input], log_probs=[], dec_state=decoder_hidden, input_feed=None, num_non_words=1, enc_attn_weights=[], use_coverage=self.use_coverage, states_for_tree=[decoder_hidden])]
    
            while len(hypos) > 0 and step <= self.max_decoder_step:
                n_hypos = len(hypos)
                if n_hypos < self.beam_size:
                    hypos.extend(hypos[-1] for _ in range(self.beam_size - n_hypos)) # check deep copy
                decoder_input = torch.tensor([h.tokens[-1] for h in hypos]).to(graph_node_embedding.device)
                decoder_hidden = (torch.cat([h.dec_state[0] for h in hypos], 0), torch.cat([h.dec_state[1] for h in hypos], 0))

                prediction, decoder_hidden, dec_attn_scores = self.decoder.decode_step(tgt_batch_size=self.beam_size, 
                                                        dec_single_input=decoder_input,
                                                        dec_single_state=decoder_hidden,
                                                        memory=single_graph_node_embedding,
                                                        parent_state=single_parent_state,
                                                        oov_dict=oov_dict,
                                                        enc_batch=enc_batch)
                prediction = torch.log(prediction + 1e-31)
                top_v, top_i = prediction.data.topk(self.beam_size)


                new_hypos = []
                for in_idx in range(n_hypos):
                    for out_idx in range(self.beam_size):
                        new_tok = top_i[in_idx][out_idx].item()
                        new_prob = top_v[in_idx][out_idx].item()
                        new_enc_attn_weights = dec_attn_scores[in_idx, :].unsqueeze(0).unsqueeze(0)

                        non_word = new_tok == self.vocab.get_symbol_idx(self.vocab.end_token)  # only SOS & EOS don't count
                        tmp_decoder_state = (decoder_hidden[0][in_idx, :].unsqueeze(0),
                                            decoder_hidden[1][in_idx, :].unsqueeze(0))

                        new_hypo = hypos[in_idx].create_next(token=new_tok, log_prob=new_prob,
                                                             dec_state=tmp_decoder_state, input_feed=None, non_word=non_word,
                                                             add_enc_attn_weights=new_enc_attn_weights)
                        new_hypos.append(new_hypo)

                new_hypos = sorted(new_hypos, key=lambda h: -h.avg_log_prob)[:self.beam_size]
                hypos = []
                new_complete_results, new_incomplete_results = [], []
                for nh in new_hypos:
                    length = len(nh)  # Does not count SOS and EOS
                    if nh.tokens[-1] == self.vocab.get_symbol_idx(self.vocab.end_token) :  # a complete hypothesis
                        if len(new_complete_results) < self.beam_size and min_out_len <= length <= max_out_len:
                            new_complete_results.append(nh)
                    elif len(hypos) < self.beam_size and length < max_out_len:  # an incomplete hypothesis
                        hypos.append(nh)
                    elif length == max_out_len and len(new_incomplete_results) < self.beam_size:
                        new_incomplete_results.append(nh)
                if new_complete_results:
                    results.extend(new_complete_results)
                elif new_incomplete_results:
                    backup_results.extend(new_incomplete_results)
                step += 1

            if not results:  # if no sequence ends with EOS within desired length, fallback to sequences
                results = backup_results  # that are "truncated" at the end to max_out_len
            batch_results.append(sorted(results, key=lambda h: -h.avg_log_prob)[:topk])

        ret = torch.zeros(batch_size, topk, self.max_decoder_step).long()
        states = []
        for sent_id, each in enumerate(batch_results):
            for i in range(topk):
                ids = torch.Tensor(each[i].tokens[:])[:self.max_decoder_step]
                if ids.shape[0] < self.max_decoder_step:
                    pad = torch.zeros(self.max_decoder_step - ids.shape[0])
                    ids = torch.cat((ids, pad), dim=0)
                ret[sent_id, i, :] = ids
                states.append(each[i].states_for_tree[:][:self.max_decoder_step])
        assert batch_size == 1 and topk == 1
        # for id 
        output_results = [[[]]]

        token_id_list = ret[0][0]
        states_emb_list = states[0]
        for index in range(self.max_decoder_step):
            if token_id_list[index] != self.vocab.get_symbol_idx(self.vocab.pad_token):
                output_results[0][0].append(BeamSearchNode(hiddenstate=states_emb_list[index],
                                                     enc_attn_weights_average=None,
                                                     previousNode=None,
                                                     wordId=token_id_list[index],
                                                     logProb=0,
                                                     length=-1))

        return output_results





from __future__ import absolute_import
from torch import nn


class ClassifierBase(nn.Module):

    def __init__(self):
        super(ClassifierBase, self).__init__()

    def forward(self, node_emb):
        raise NotImplementedError()


class ClassifierLayerBase(nn.Module):

    def __init__(self):
        super(ClassifierLayerBase, self).__init__()

    def forward(self, node_emb):
        raise NotImplementedError()


class NodeClassifierBase(ClassifierBase):

    def __init__(self):
        super(NodeClassifierBase, self).__init__()

    def forward(self, node_emb, node_idx=None):
        raise NotImplementedError()


class NodeClassifierLayerBase(ClassifierLayerBase):

    def __init__(self):
        super(NodeClassifierLayerBase, self).__init__()

    def forward(self, node_emb, node_idx=None):

        raise NotImplementedError()


class KGCompletionBase(nn.Module):

    def __init__(self):
        super(KGCompletionBase, self).__init__()

    def forward(self, input_graph, e1_emb, rel_emb, all_node_emb, multi_label=None):
        # Cannot inherit from the base class ClassifierLayerBase and
        # ClassifierBase because the forward(...) arguments are inconsistent.
        # rel_emb cannot be `None`
        raise NotImplementedError()


class KGCompletionLayerBase(nn.Module):

    def __init__(self):
        super(KGCompletionLayerBase, self).__init__()

    def forward(self, e1_emb, rel_emb, all_node_emb, multi_label=None):
        raise NotImplementedError()


class LinkPredictionBase(ClassifierBase):

    def __init__(self):
        super(LinkPredictionBase, self).__init__()

    def forward(self, node_emb, node_idx=None):
        raise NotImplementedError()


class LinkPredictionLayerBase(ClassifierLayerBase):

    def __init__(self):
        super(LinkPredictionLayerBase, self).__init__()

    def forward(self, node_emb, node_idx=None):

        raise NotImplementedError()


class GraphClassifierBase(ClassifierBase):

    def __init__(self):
        super(GraphClassifierBase, self).__init__()

    def forward(self, graph_emb):
        raise NotImplementedError()


class GraphClassifierLayerBase(ClassifierLayerBase):

    def __init__(self):
        super(GraphClassifierLayerBase, self).__init__()

    def forward(self, graph_emb):

        raise NotImplementedError()

class PoolingBase(nn.Module):

    def __init__(self):
        super(PoolingBase, self).__init__()

    def forward(self, graph, feat):

        raise NotImplementedError()



import torch.nn as nn

from .base import GeneralLossBase


class GeneralLoss(GeneralLossBase):
    r"""
        This general loss are backended on the pytorch loss function.
        The detailed decription for each loss function can be found at:
            `pytorch loss function <https://pytorch.org/docs/stable/nn.html#loss-functions>`

    Parameters
    ----------
    loss_type: str
        the loss function to select (``NLL``,``BCEWithLogits``, ``MultiLabelMargin``,``SoftMargin``,``CrossEntropy`` )

        `NLL loss<https://pytorch.org/docs/stable/_modules/torch/nn/modules/loss.html#NLLLoss>` measures
        the negative log likelihood loss. It is useful to train a classification problem with C classes.

        `BCEWithLogits loss<https://pytorch.org/docs/stable/_modules/torch/nn/modules/loss.html#BCEWithLogitsLoss>` combines a
        `Sigmoid` layer and the `BCELoss` in one single class. This version is more numerically stable than using a plain `Sigmoid`
        followed by a `BCELoss` as, by combining the operations into one layer, we take advantage of the log-sum-exp trick for numerical stability.

        `BCE Loss<https://pytorch.org/docs/stable/_modules/torch/nn/modules/loss.html#BCELoss>` creates a criterion that measures
        the Binary Cross Entropy between the target and the output.

        `MultiLabelMargin loss<https://pytorch.org/docs/stable/_modules/torch/nn/modules/loss.html#MultiLabelMarginLoss>` creates a
        criterion that optimizes a multi-class multi-classification hinge loss (margin-based loss) between
        input :math:`x` (a 2D mini-batch `Tensor`) and output :math:`y` (which is a 2D `Tensor` of target class indices).

        `SoftMargin loss<https://pytorch.org/docs/stable/_modules/torch/nn/modules/loss.html#SoftMarginLoss>` creates a criterion
        that optimizes a two-class classification logistic loss between input tensor :math:`x` and target tensor :math:`y` (containing 1 or -1).

        `CrossEntropy loss<https://pytorch.org/docs/stable/_modules/torch/nn/modules/loss.html#CrossEntropyLoss> `combines
        pytorch function `nn.LogSoftmax` and `nn.NLLLoss` in one single class. It is useful when training a classification
        problem with `C` classes.

    weight: Tensor, optional
        a manual rescaling weight given to the loss of each batch element. If given, has to be a Tensor of size `nbatch`.
        This parameter is not suitable for ``SoftMargin`` loss functions.
    size_average: bool, optional
        By default,the losses are averaged over each loss element in the batch. Note that for some losses, there are
        multiple elements per sample. If the field :attr:`size_average` is set to ``False``, the losses are instead summed
        for each minibatch. Ignored when reduce is ``False``. Default: ``True``.
    reduce: bool, optional
        By default, the losses are averaged or summed over observations for each minibatch depending on :attr:`size_average`.
        When :attr:`reduce` is ``False``, returns a loss per batch element instead and ignores :attr:`size_average`. Default: ``True``
    reduction: string, optional
        Specifies the reduction to apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
          ``'mean'``: the sum of the output will be divided by the number of elements in the output, ``'sum'``: the output will be summed.
        Note: :attr:`size_average` and :attr:`reduce` are in the process of being deprecated, and in the meantime, specifying either of
        those two args will override :attr:`reduction`. Default: ``'mean'``
    pos_weight:Tensor, optional
        A weight of positive examples. Must be a vector with length equal to the number of classes. This paramter is only suitable for
        ``BCEWithLogits`` loss function.
    ignore_index: int, optional
        Specifies a target value that is ignored and does not contribute to the input gradient. When :attr:`size_average` is
        ``True``, the loss is averaged over non-ignored targets. This paramter is only suitable for ``CrossEntropy`` loss function.
    """

    def __init__(self, loss_type, weight=None, size_average=None, ignore_index=-100,
                 reduce=None, reduction='mean', pos_weight=None):
        super(GeneralLoss, self).__init__()
        if loss_type == 'NLL':
            self.loss_function = nn.NLLLoss(weight, size_average, ignore_index, reduce, reduction)
        if loss_type == 'BCEWithLogits':
            self.loss_function = nn.BCEWithLogitsLoss(weight, size_average, reduce, reduction, pos_weight)
        if loss_type == 'MultiLabelMargin':
            self.loss_function = nn.MultiLabelMarginLoss(size_average, reduce, reduction)
        if loss_type == 'SoftMargin':
            self.loss_function = nn.SoftMarginLoss(size_average, reduce, reduction)
        if loss_type == 'CrossEntropy':
            self.loss_function = nn.CrossEntropyLoss(weight, size_average, ignore_index, reduce, reduction)
        if loss_type == 'BCE':
            self.loss_function = nn.CrossEntropyLoss(weight, size_average, reduce, reduction)

    def forward(self, input, target):
        r"""
        Compute the loss.

        Parameters
        ----------
        NLL loss:
             Input: tensor.
               :math:`(N, C)` where `C = number of classes`, or :math:`(N, C, d_1, d_2, ..., d_K)` with :math:`K \geq 1`
                in the case of `K`-dimensional loss.
             Target: tensor.
               :math:`(N)` where each value is :math:`0 \leq \text{targets}[i] \leq C-1`, or :math:`(N, d_1, d_2, ..., d_K)`
                with :math:`K \geq 1` in the case of K-dimensional loss.
             Output: scalar.
               If :attr:`reduction` is ``'none'``, then the same size as the target: :math:`(N)`, or :math:`(N, d_1, d_2, ..., d_K)`
               with :math:`K \geq 1` in the case of K-dimensional loss.

        BCE/BCEWithLogits loss:
             Input: Tensor.
               :math:`(N, *)` where :math:`*` means, any number of additional dimensions
             Target: Tensor.
               :math:`(N, *)`, same shape as the input
             Output: scalar.
               If :attr:`reduction` is ``'none'``, then :math:`(N, *)`, same shape as input.

        MultiLabelMargin loss:
             Input: Tensor.
               :math:`(C)` or :math:`(N, C)` where `N` is the batch size and `C` is the number of classes.
             Target: Tensor.
               :math:`(C)` or :math:`(N, C)`, label targets padded by -1 ensuring same shape as the input.
             Output: Scalar.
                If :attr:`reduction` is ``'none'``, then :math:`(N)`.

        SoftMargin loss:
             Input: Tensor.
               :math:`(*)` where :math:`*` means, any number of additional dimensions
             Target: Tensor.
               :math:`(*)`, same shape as the input
             Output: scalar.
               If :attr:`reduction` is ``'none'``, then same shape as the input

        CrossEntropy:
             Input: Tensor.
               :math:`(N, C)` where `C = number of classes`, or :math:`(N, C, d_1, d_2, ..., d_K)` with :math:`K \geq 1`
               in the case of `K`-dimensional loss.
             Target: Tensor.
               :math:`(N)` where each value is :math:`0 \leq \text{targets}[i] \leq C-1`, or :math:`(N, d_1, d_2, ..., d_K)`
               with :math:`K \geq 1` in the case of K-dimensional loss.
            Output: scalar.
               If :attr:`reduction` is ``'none'``, then the same size as the target: :math:`(N)`, or :math:`(N, d_1, d_2, ..., d_K)`
               with :math:`K \geq 1` in the case of K-dimensional loss.
        """

        return self.loss_function(input, target)



#!/usr/bin/env python

# bleu_scorer.py
# David Chiang <chiang@isi.edu>

# Copyright (c) 2004-2006 University of Maryland. All rights
# reserved. Do not redistribute without permission from the
# author. Not for commercial use.

# Modified by:
# Hao Fang <hfang@uw.edu>
# Tsung-Yi Lin <tl483@cornell.edu>

'''Provides:
cook_refs(refs, n=4): Transform a list of reference sentences as strings into a form usable by cook_test().
cook_test(test, refs, n=4): Transform a test sentence as a string (together with the cooked reference sentences) into a form usable by score_cooked().
'''
from builtins import range, dict

import copy
import sys, math, re
from collections import defaultdict

def precook(s, n=4, out=False):
    """Takes a string as input and returns an object that can be given to
    either cook_refs or cook_test. This is optional: cook_refs and cook_test
    can take string arguments as well."""
    words = s.split()
    counts = defaultdict(int)
    for k in range(1,n+1):
        for i in range(len(words)-k+1):
            ngram = tuple(words[i:i+k])
            counts[ngram] += 1
    return (len(words), counts)

def cook_refs(refs, eff=None, n=4): ## lhuang: oracle will call with "average"
    '''Takes a list of reference sentences for a single segment
    and returns an object that encapsulates everything that BLEU
    needs to know about them.'''

    reflen = []
    maxcounts = dict()
    for ref in refs:
        rl, counts = precook(ref, n)
        reflen.append(rl)
        for (ngram,count) in counts.items():
            maxcounts[ngram] = max(maxcounts.get(ngram,0), count)

    # Calculate effective reference sentence length.
    if eff == "shortest":
        reflen = min(reflen)
    elif eff == "average":
        reflen = float(sum(reflen))/len(reflen)

    ## lhuang: N.B.: leave reflen computaiton to the very end!!

    ## lhuang: N.B.: in case of "closest", keep a list of reflens!! (bad design)

    return (reflen, maxcounts)

def cook_test(test, refs, eff=None, n=4):
    '''Takes a test sentence and returns an object that
    encapsulates everything that BLEU needs to know about it.'''

    reflen, refmaxcounts = refs
    testlen, counts = precook(test, n, True)

    result = dict()

    # Calculate effective reference sentence length.

    if eff == "closest":
        result["reflen"] = min((abs(l-testlen), l) for l in reflen)[1]
    else: ## i.e., "average" or "shortest" or None
        result["reflen"] = reflen

    result["testlen"] = testlen

    result["guess"] = [max(0,testlen-k+1) for k in range(1,n+1)]

    result['correct'] = [0]*n
    for (ngram, count) in counts.items():
        result["correct"][len(ngram)-1] += min(refmaxcounts.get(ngram,0), count)

    return result

class BleuScorer(object):
    """Bleu scorer.
    """

    __slots__ = "n", "crefs", "ctest", "_score", "_ratio", "_testlen", "_reflen", "special_reflen"
    # special_reflen is used in oracle (proportional effective ref len for a node).

    def copy(self):
        ''' copy the refs.'''
        new = BleuScorer(n=self.n)
        new.ctest = copy.copy(self.ctest)
        new.crefs = copy.copy(self.crefs)
        new._score = None
        return new

    def __init__(self, test=None, refs=None, n=4, special_reflen=None):
        ''' singular instance '''

        self.n = n
        self.crefs = []
        self.ctest = []
        self.cook_append(test, refs)
        self.special_reflen = special_reflen

    def cook_append(self, test, refs):
        '''called by constructor and __iadd__ to avoid creating new instances.'''

        if refs is not None:
            self.crefs.append(cook_refs(refs))
            if test is not None:
                cooked_test = cook_test(test, self.crefs[-1])
                self.ctest.append(cooked_test) ## N.B.: -1
            else:
                self.ctest.append(None) # lens of crefs and ctest have to match

        self._score = None ## need to recompute

    def ratio(self, option=None):
        self.compute_score(option=option)
        return self._ratio

    def score_ratio(self, option=None):
        '''return (bleu_tool, len_ratio) pair'''
        return (self.fscore(option=option), self.ratio(option=option))

    def score_ratio_str(self, option=None):
        return "%.4f (%.2f)" % self.score_ratio(option)

    def reflen(self, option=None):
        self.compute_score(option=option)
        return self._reflen

    def testlen(self, option=None):
        self.compute_score(option=option)
        return self._testlen

    def retest(self, new_test):
        if type(new_test) is str:
            new_test = [new_test]
        assert len(new_test) == len(self.crefs), new_test
        self.ctest = []
        for t, rs in zip(new_test, self.crefs):
            self.ctest.append(cook_test(t, rs))
        self._score = None

        return self

    def rescore(self, new_test):
        ''' replace test(s) with new test(s), and returns the new score.'''

        return self.retest(new_test).compute_score()

    def size(self):
        assert len(self.crefs) == len(self.ctest), "refs/test mismatch! %d<>%d" % (len(self.crefs), len(self.ctest))
        return len(self.crefs)

    def __iadd__(self, other):
        '''add an instance (e.g., from another sentence).'''

        if type(other) is tuple:
            ## avoid creating new BleuScorer instances
            self.cook_append(other[0], other[1])
        else:
            assert self.compatible(other), "incompatible BLEUs."
            self.ctest.extend(other.ctest)
            self.crefs.extend(other.crefs)
            self._score = None ## need to recompute

        return self

    def compatible(self, other):
        return isinstance(other, BleuScorer) and self.n == other.n

    def single_reflen(self, option="average"):
        return self._single_reflen(self.crefs[0][0], option)

    def _single_reflen(self, reflens, option=None, testlen=None):

        if option == "shortest":
            reflen = min(reflens)
        elif option == "average":
            reflen = float(sum(reflens))/len(reflens)
        elif option == "closest":
            reflen = min((abs(l-testlen), l) for l in reflens)[1]
        else:
            assert False, "unsupported reflen option %s" % option

        return reflen

    def recompute_score(self, option=None, verbose=0):
        self._score = None
        return self.compute_score(option, verbose)

    def compute_score(self, option=None, verbose=0):
        n = self.n
        small = 1e-9
        tiny = 1e-15 ## so that if guess is 0 still return 0
        bleu_list = [[] for _ in range(n)]

        if self._score is not None:
            return self._score

        if option is None:
            option = "average" if len(self.crefs) == 1 else "closest"

        self._testlen = 0
        self._reflen = 0
        totalcomps = {'testlen':0, 'reflen':0, 'guess':[0]*n, 'correct':[0]*n}

        # for each sentence
        for comps in self.ctest:
            testlen = comps['testlen']
            self._testlen += testlen

            if self.special_reflen is None: ## need computation
                reflen = self._single_reflen(comps['reflen'], option, testlen)
            else:
                reflen = self.special_reflen

            self._reflen += reflen

            for key in ['guess','correct']:
                for k in range(n):
                    totalcomps[key][k] += comps[key][k]

            # append per image bleu_tool score
            bleu = 1.
            for k in range(n):
                bleu *= (float(comps['correct'][k]) + tiny) \
                        /(float(comps['guess'][k]) + small)
                bleu_list[k].append(bleu ** (1./(k+1)))
            ratio = (testlen + tiny) / (reflen + small) ## N.B.: avoid zero division
            if ratio < 1:
                for k in range(n):
                    bleu_list[k][-1] *= math.exp(1 - 1/ratio)

            if verbose > 1:
                print(comps, reflen)

        totalcomps['reflen'] = self._reflen
        totalcomps['testlen'] = self._testlen

        bleus = []
        bleu = 1.
        for k in range(n):
            bleu *= float(totalcomps['correct'][k] + tiny) \
                    / (totalcomps['guess'][k] + small)
            bleus.append(bleu ** (1./(k+1)))
        ratio = (self._testlen + tiny) / (self._reflen + small) ## N.B.: avoid zero division
        if ratio < 1:
            for k in range(n):
                bleus[k] *= math.exp(1 - 1/ratio)

        if verbose > 0:
            print(totalcomps)
            print("ratio:", ratio)

        self._score = bleus
        return self._score, bleu_list



from ..base import KGCompletionBase
from .DistMultLayer import DistMultLayer
from .....data.data import GraphData


class DistMult(KGCompletionBase):
    r"""Specific class for knowledge graph completion task.

    DistMult from paper `Embedding entities and relations for learning and
    inference in knowledge bases <https://arxiv.org/pdf/1412.6575.pdf>`__.

    Parameters
    ----------
    input_dropout: float
        Dropout for node_emb and rel_emb. Default: 0.0

    loss_name: str
        The loss type selected fot the KG completion task. Default: `'BCELoss'`
    """

    def __init__(self,
                 input_dropout=0.0,
                 loss_name='BCELoss'):
        super(DistMult, self).__init__()
        self.loss_name = loss_name
        self.classifier = DistMultLayer(input_dropout, loss_name)

    def forward(self, input_graph: GraphData, e1_emb, rel_emb, all_node_emb, multi_label=None):
        r"""
        Forward functions to compute the logits tensor for kg completion.

        Parameters
        ----------

        input graph : GraphData
            The tensors stored in the node feature field named "node_emb" and
            "rel_emb" in the input_graph are used for knowledge graph completion.

        e1_emb : tensor [B, H]
            The selected entity_1 embeddings of a batch.
            B: batch size
            H: length of the node embeddings (entity embeddings)

        rel_emb : tensor [B, H]
            The selected relation embeddings of a batch.
            B: batch size
            H: length of the edge embeddings (relation embeddings)

        all_node_emb :  torch.nn.modules.sparse.Embedding [N, H]
            All node embeddings.
            N: number of nodes in the whole KG graph
            H: length of the node embeddings (entity embeddings)

        multi_label: tensor [B, N]
            multi_label is a binary matrix. Each element can be equal to 1 for true label
            and 0 for false label (or 1 for true label, -1 for false label).
            multi_label[i] represents a multi-label of a given head-rel pair.
            B is the batch size.
            N: number of nodes in the whole KG graph.

        Returns
        ---------

        output_graph : GraphData
                      The computed logit tensor for each nodes in the graph are stored
                      in the node feature field named "node_logits".
                      logit tensor shape is: [num_class]
        """

        if multi_label is None:
            input_graph.graph_attributes['logits'] = self.classifier(e1_emb,
                                                                     rel_emb,
                                                                     all_node_emb)  # [B, N]
        else:
            input_graph.graph_attributes['logits'], input_graph.graph_attributes['p_score'], \
            input_graph.graph_attributes['n_score'] = self.classifier(e1_emb,
                                                                      rel_emb,
                                                                      all_node_emb,
                                                                      multi_label)
            # input_graph.graph_attributes['p_score']: [L_p]
            # input_graph.graph_attributes['n_score']: [L_n]
            # L_p + L_n == B * N

        return input_graph


import collections
from torch import nn

from ..base import GraphClassifierLayerBase, GraphClassifierBase
from .avg_pooling import AvgPooling
from .max_pooling import MaxPooling


class FeedForwardNN(GraphClassifierBase):
    r"""FeedForwardNN class for graph classification task.

    Parameters
    ----------
    input_size : int
        The dimension of input graph embeddings.
    num_class : int
        The number of classes for classification.
    hidden_size : list of int
        Hidden size per NN layer.
    activation: nn.Module, optional
        The activation function, default: `nn.ReLU()`.
    """
    def __init__(self,
                input_size,
                num_class,
                hidden_size,
                activation=None,
                graph_pool_type='max_pool',
                **kwargs):
        super(FeedForwardNN, self).__init__()

        if not activation:
            activation = nn.ReLU()

        if graph_pool_type == 'avg_pool':
            self.graph_pool = AvgPooling()
        elif graph_pool_type == 'max_pool':
            self.graph_pool = MaxPooling(**kwargs)
        else:
            raise RuntimeError('Unknown graph pooling type: {}'.format(graph_pool_type))

        self.classifier = FeedForwardNNLayer(input_size, num_class, hidden_size, activation)

    def forward(self, graph):
        r"""Compute the logits tensor for graph classification.

        Parameters
        ----------
        graph : GraphData
            The graph data containing graph embeddings.

        Returns
        -------
        list of GraphData
            The output graph data containing logits tensor for graph classification.
        """
        graph_emb = self.graph_pool(graph, 'node_emb')
        logits = self.classifier(graph_emb)
        graph.graph_attributes['logits'] = logits

        return graph

class FeedForwardNNLayer(GraphClassifierLayerBase):
    r"""FeedForwardNNLayer class for graph classification task.

    Parameters
    ----------
    input_size : int
        The dimension of input graph embeddings.
    num_class : int
        The number of classes for classification.
    hidden_size : list of int
        Hidden size per NN layer.
    activation: nn.Module, optional
        The activation function, default: `nn.ReLU()`.
    """
    def __init__(self, input_size, num_class, hidden_size, activation=None):
        super(FeedForwardNNLayer, self).__init__()

        if not activation:
            activation = nn.ReLU()

        # build the linear module list
        module_seq = []

        for layer_idx in range(len(hidden_size)):
            if layer_idx == 0:
                module_seq.append(('fc' + str(layer_idx),
                    nn.Linear(input_size, hidden_size[layer_idx])))
            else:
                module_seq.append(('fc' + str(layer_idx),
                    nn.Linear(hidden_size[layer_idx - 1], self.hidden_size[layer_idx])))

            module_seq.append(('activate' + str(layer_idx), activation))

        module_seq.append(('fc_end', nn.Linear(hidden_size[-1], num_class)))

        self.classifier = nn.Sequential(collections.OrderedDict(module_seq))

    def forward(self, graph_emb):
        r"""Compute the logits tensor for graph classification.

        Parameters
        ----------
        graph_emb : torch.Tensor
            The input graph embeddings.

        Returns
        -------
        torch.Tensor
            The output logits tensor for graph classification.
        """
        return self.classifier(graph_emb)



"""
graph4nlp.datasets module contains various common datasets implemented based on graph4nlp.data.dataset.
"""

from .geo import GeoDatasetForTree
from .jobs import JobsDataset, JobsDatasetForTree
from .mawps import MawpsDatasetForTree
from .squad import SQuADDataset
from .trec import TrecDataset
from .mathqa import MathQADatasetForTree